import asyncio
import os
from datetime import datetime
from unittest import TestCase

from goodwe.dt import DT
from goodwe.exceptions import RequestFailedException, RequestRejectedException
from goodwe.modbus import ILLEGAL_DATA_ADDRESS
from goodwe.protocol import ProtocolCommand, ProtocolResponse


class DtMock(TestCase, DT):

    def __init__(self, methodName='runTest', port=8899):
        TestCase.__init__(self, methodName)
        DT.__init__(self, "localhost", port)
        self.sensor_map = {s.id_: s for s in self.sensors()}
        self._mock_responses = {}

    def mock_response(self, command: ProtocolCommand, filename: str):
        self._mock_responses[command] = filename

    async def _read_from_socket(self, command: ProtocolCommand) -> ProtocolResponse:
        """Mock UDP communication"""
        root_dir = os.path.dirname(os.path.abspath(__file__))
        filename = self._mock_responses.get(command)
        if filename is not None:
            if ILLEGAL_DATA_ADDRESS == filename:
                raise RequestRejectedException(ILLEGAL_DATA_ADDRESS)
            if 'NO RESPONSE' == filename:
                raise RequestFailedException()
            with open(root_dir + '/sample/dt/' + filename, 'r') as f:
                response = bytes.fromhex(f.read())
                if not command.validator(response):
                    raise RequestFailedException
                return ProtocolResponse(response, command)
        else:
            self.request = command.request
            return ProtocolResponse(bytes.fromhex("aa557f00010203040506070809"), command)

    def assertSensor(self, sensor_name, expected_value, expected_unit, data):
        self.assertEqual(expected_value, data.get(sensor_name))
        sensor = self.sensor_map.get(sensor_name);
        self.assertEqual(expected_unit, sensor.unit)
        self.sensor_map.pop(sensor_name)

    @classmethod
    def setUpClass(cls):
        cls.loop = asyncio.get_event_loop()


class GW6000_DT_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_RUNNING_DATA, 'GW6000-DT_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW6000_DT_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(42, len(data))

        self.sensor_map = {s.id_: s for s in self.sensors()}

        self.assertSensor('timestamp', datetime.strptime('2021-08-31 12:03:02', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 320.8, 'V', data)
        self.assertSensor('ipv1', 3.1, 'A', data)
        self.assertSensor('ppv1', 994, 'W', data)
        self.assertSensor('vpv2', 324.1, 'V', data)
        self.assertSensor('ipv2', 3.2, 'A', data)
        self.assertSensor('ppv2', 1037, 'W', data)
        self.assertSensor('ppv', 2031, 'W', data)
        self.assertSensor('vline1', 0, 'V', data)
        self.assertSensor('vline2', 0, 'V', data)
        self.assertSensor('vline3', 0, 'V', data)
        self.assertSensor('vgrid1', 225.6, 'V', data)
        self.assertSensor('vgrid2', 229.7, 'V', data)
        self.assertSensor('vgrid3', 231.0, 'V', data)
        self.assertSensor('igrid1', 2.7, 'A', data)
        self.assertSensor('igrid2', 2.6, 'A', data)
        self.assertSensor('igrid3', 2.7, 'A', data)
        self.assertSensor('fgrid1', 49.98, 'Hz', data)
        self.assertSensor('fgrid2', 49.98, 'Hz', data)
        self.assertSensor('fgrid3', 49.98, 'Hz', data)
        self.assertSensor('pgrid1', 609, 'W', data)
        self.assertSensor('pgrid2', 597, 'W', data)
        self.assertSensor('pgrid3', 624, 'W', data)
        self.assertSensor('total_inverter_power', 1835, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", -1, "VA", data),
        self.assertSensor("reactive_power", -1, "var", data),
        self.assertSensor("power_factor", 0.0, "", data),
        self.assertSensor('temperature', 41.3, 'C', data)
        self.assertSensor('e_day', 6.0, 'kWh', data)
        self.assertSensor('e_total', 13350.2, 'kWh', data)
        self.assertSensor('h_total', 8451, 'h', data)
        self.assertSensor('safety_country', 20, '', data)
        self.assertSensor('safety_country_label', 'NL-A', '', data)
        self.assertSensor('funbit', 0, '', data)
        self.assertSensor('vbus', 601.2, 'V', data)
        self.assertSensor('vnbus', 305.4, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)

        self.assertFalse(self.sensor_map, f"Some sensors were not tested {self.sensor_map}")

    def test_GW6000_DT_setting(self):
        self.assertEqual(8, len(self.settings()))
        settings = {s.id_: s for s in self.settings()}
        self.assertEqual('Timestamp', type(settings.get("time")).__name__)
        self.assertEqual('Integer', type(settings.get("grid_export")).__name__)
        self.assertEqual('Integer', type(settings.get("grid_export_limit")).__name__)

    def test_GW6000_DT_read_setting(self):
        self.loop.run_until_complete(self.read_setting('shadow_scan'))
        self.assertEqual('7f039d8600014051', self.request.hex())

    def test_GW6000_DT_write_setting(self):
        self.loop.run_until_complete(self.write_setting('shadow_scan', 1))
        self.assertEqual('7f069d8600018c51', self.request.hex())


class GW8K_DT_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW8K-DT_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW8K-DT_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW8K_DT_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW8K-DT', self.model_name)
        self.assertEqual('00000DTS00000000', self.serial_number)
        self.assertEqual(1010, self.dsp1_version)
        self.assertEqual(1010, self.dsp2_version)
        self.assertEqual(728, self.dsp_svn_version)
        self.assertEqual(8, self.arm_version)
        self.assertEqual(49, self.arm_svn_version)
        self.assertEqual('1010.1010.08', self.firmware)

    def test_GW8K_DT_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(42, len(data))

        self.assertSensor('timestamp', datetime.strptime('2021-08-24 16:43:27', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 275.5, 'V', data)
        self.assertSensor('ipv1', 0.6, 'A', data)
        self.assertSensor('ppv1', 165, 'W', data)
        self.assertSensor('vpv2', 510.8, 'V', data)
        self.assertSensor('ipv2', 0.8, 'A', data)
        self.assertSensor('ppv2', 409, 'W', data)
        self.assertSensor('ppv', 574, 'W', data)
        self.assertSensor('vline1', 413.7, 'V', data)
        self.assertSensor('vline2', 413.0, 'V', data)
        self.assertSensor('vline3', 408.0, 'V', data)
        self.assertSensor('vgrid1', 237.2, 'V', data)
        self.assertSensor('vgrid2', 240.5, 'V', data)
        self.assertSensor('vgrid3', 235.2, 'V', data)
        self.assertSensor('igrid1', 1.0, 'A', data)
        self.assertSensor('igrid2', 1.0, 'A', data)
        self.assertSensor('igrid3', 1.0, 'A', data)
        self.assertSensor('fgrid1', 50.08, 'Hz', data)
        self.assertSensor('fgrid2', 50.04, 'Hz', data)
        self.assertSensor('fgrid3', 50.0, 'Hz', data)
        self.assertSensor('pgrid1', 237, 'W', data)
        self.assertSensor('pgrid2', 240, 'W', data)
        self.assertSensor('pgrid3', 235, 'W', data)
        self.assertSensor('total_inverter_power', 643, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", 0, "VA", data),
        self.assertSensor("reactive_power", 0, "var", data),
        self.assertSensor("power_factor", 0.0, "", data),
        self.assertSensor('temperature', 45.3, 'C', data)
        self.assertSensor('e_day', None, 'kWh', data)
        self.assertSensor('e_total', None, 'kWh', data)
        self.assertSensor('h_total', 0, 'h', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('funbit', 512, '', data)
        self.assertSensor('vbus', 624.2, 'V', data)
        self.assertSensor('vnbus', 316.8, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)

    def test_get_grid_export_limit(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.get_grid_export_limit())
        self.assertEqual('7f039d900001a195', self.request.hex())

    def test_set_grid_export_limit(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.set_grid_export_limit(100))
        self.assertEqual('7f069d900064adbe', self.request.hex())


class GW5000D_NS_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'Mock_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW5000D-NS_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW5000D_NS_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(32, len(data))

        self.assertSensor('timestamp', datetime.strptime('2021-09-06 06:56:01', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 224.4, 'V', data)
        self.assertSensor('ipv1', 0.0, 'A', data)
        self.assertSensor('ppv1', 0, 'W', data)
        self.assertSensor('vpv2', 291.8, 'V', data)
        self.assertSensor('ipv2', 0, 'A', data)
        self.assertSensor('ppv2', 0, 'W', data)
        self.assertSensor('ppv', 0, 'W', data)
        self.assertSensor('vline1', 0, 'V', data)
        self.assertSensor('vgrid1', 240.5, 'V', data)
        self.assertSensor('igrid1', 0.0, 'A', data)
        self.assertSensor('fgrid1', 49.97, 'Hz', data)
        self.assertSensor('pgrid1', 0, 'W', data)
        self.assertSensor('total_inverter_power', 0, 'W', data)
        self.assertSensor('work_mode', 0, '', data)
        self.assertSensor('work_mode_label', 'Wait Mode', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", -1, "VA", data),
        self.assertSensor("reactive_power", -1, "var", data),
        self.assertSensor("power_factor", -0.001, "", data),
        self.assertSensor('temperature', 1.4, 'C', data)
        self.assertSensor('e_day', 0, 'kWh', data)
        self.assertSensor('e_total', 881.7, 'kWh', data)
        self.assertSensor('h_total', 955, 'h', data)
        self.assertSensor('safety_country', 73, '', data)
        self.assertSensor('safety_country_label', 'Australia Victoria', '', data)
        self.assertSensor('funbit', 2400, '', data)
        self.assertSensor('vbus', 291.7, 'V', data)
        self.assertSensor('vnbus', 0, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)

    def test_get_grid_export_limit(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.get_grid_export_limit())
        self.assertEqual('7f039d8800026193', self.request.hex())

    def test_set_grid_export_limit(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.set_grid_export_limit(5000))
        self.assertEqual('7f109d88000204000013889d80', self.request.hex())


class GW5000_MS_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW5000-MS_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW5000-MS_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW6000_MS_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW5000-MS', self.model_name)
        self.assertEqual('00000MSU00000000', self.serial_number)
        self.assertEqual(12, self.dsp1_version)
        self.assertEqual(12, self.dsp2_version)
        self.assertEqual(65535, self.dsp_svn_version)
        self.assertEqual(16, self.arm_version)
        self.assertEqual(271, self.arm_svn_version)
        self.assertEqual('12.12.10', self.firmware)

    def test_GW5000_MS_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(35, len(data))

        self.assertSensor('timestamp', datetime.strptime('2021-10-15 09:03:12', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 319.6, 'V', data)
        self.assertSensor('ipv1', 0.2, 'A', data)
        self.assertSensor('ppv1', 64, 'W', data)
        self.assertSensor('vpv2', 148.0, 'V', data)
        self.assertSensor('ipv2', 0.3, 'A', data)
        self.assertSensor('ppv2', 44, 'W', data)
        self.assertSensor('vpv3', 143.2, 'V', data)
        self.assertSensor('ipv3', 0.4, 'A', data)
        self.assertSensor('ppv3', 57, 'W', data)
        self.assertSensor('ppv', 165, 'W', data)
        self.assertSensor('vline1', 0, 'V', data)
        self.assertSensor('vgrid1', 240.1, 'V', data)
        self.assertSensor('igrid1', 0.9, 'A', data)
        self.assertSensor('fgrid1', 49.98, 'Hz', data)
        self.assertSensor('pgrid1', 216, 'W', data)
        self.assertSensor('total_inverter_power', 295, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", -1, "VA", data),
        self.assertSensor("reactive_power", -1, "var", data),
        self.assertSensor("power_factor", -0.001, "", data),
        self.assertSensor('temperature', 10.7, 'C', data)
        self.assertSensor('e_day', 0.4, 'kWh', data)
        self.assertSensor('e_total', 6.8, 'kWh', data)
        self.assertSensor('h_total', 7, 'h', data)
        self.assertSensor('safety_country', 73, '', data)
        self.assertSensor('safety_country_label', 'Australia Victoria', '', data)
        self.assertSensor('funbit', 2384, '', data)
        self.assertSensor('vbus', 393.9, 'V', data)
        self.assertSensor('vnbus', 0, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)


class GW10K_MS_30_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW10K-MS-30_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW10K-MS-30_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW10K_MS_30_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual(None, self.model_name)
        self.assertEqual('5010KMSC000W0000', self.serial_number)
        self.assertEqual(0, self.dsp1_version)
        self.assertEqual(0, self.dsp2_version)
        self.assertEqual(504, self.dsp_svn_version)
        self.assertEqual(2, self.arm_version)
        self.assertEqual(13, self.arm_svn_version)
        self.assertEqual('0.0.02', self.firmware)

    def test_GW10K_MS_30_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(35, len(data))

        self.assertSensor('timestamp', datetime.strptime('2024-01-09 22:08:20', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 0.0, 'V', data)
        self.assertSensor('ipv1', 0.0, 'A', data)
        self.assertSensor('ppv1', 0, 'W', data)
        self.assertSensor('vpv2', 0.0, 'V', data)
        self.assertSensor('ipv2', 0.0, 'A', data)
        self.assertSensor('ppv2', 0, 'W', data)
        self.assertSensor('vpv3', 0.0, 'V', data)
        self.assertSensor('ipv3', 0.0, 'A', data)
        self.assertSensor('ppv3', 0, 'W', data)
        self.assertSensor('ppv', 0, 'W', data)
        self.assertSensor('vline1', 0.0, 'V', data)
        self.assertSensor('vgrid1', 236.2, 'V', data)
        self.assertSensor('igrid1', 0.0, 'A', data)
        self.assertSensor('fgrid1', 50.0, 'Hz', data)
        self.assertSensor('pgrid1', 0, 'W', data)
        self.assertSensor('total_inverter_power', 0, 'W', data)
        self.assertSensor('work_mode', 0, '', data)
        self.assertSensor('work_mode_label', 'Wait Mode', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", 0, "VA", data),
        self.assertSensor("reactive_power", 0, "var", data),
        self.assertSensor("power_factor", 0.0, "", data),
        self.assertSensor('temperature', 24.3, 'C', data)
        self.assertSensor('e_day', 71.8, 'kWh', data)
        self.assertSensor('e_total', 3433.4, 'kWh', data)
        self.assertSensor('h_total', 971, 'h', data)
        self.assertSensor('safety_country', 9, '', data)
        self.assertSensor('safety_country_label', 'Australia A', '', data)
        self.assertSensor('funbit', 0, '', data)
        self.assertSensor('vbus', 5.1, 'V', data)
        self.assertSensor('vnbus', 0.0, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)


class GW10K_MS_TCP_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName, 502)
        self.mock_response(self._READ_RUNNING_DATA, 'GW10K-MS-30_tcp_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW10K_MS_TCP_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(42, len(data))

        self.assertSensor('timestamp', datetime.strptime('2024-06-02 09:07:17', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 400.6, 'V', data)
        self.assertSensor('ipv1', 6.9, 'A', data)
        self.assertSensor('ppv1', 2764, 'W', data)
        self.assertSensor('vpv2', 364.0, 'V', data)
        self.assertSensor('ipv2', 3.6, 'A', data)
        self.assertSensor('ppv2', 1310, 'W', data)
        self.assertSensor('ppv', 6143, 'W', data)
        self.assertSensor('vline1', 0, 'V', data)
        self.assertSensor('vline2', 0, 'V', data)
        self.assertSensor('vline3', 0, 'V', data)
        self.assertSensor('vgrid1', 241.1, 'V', data)
        self.assertSensor('vgrid2', 0, 'V', data)
        self.assertSensor('vgrid3', 0, 'V', data)
        self.assertSensor('igrid1', 24.7, 'A', data)
        self.assertSensor('igrid2', 0, 'A', data)
        self.assertSensor('igrid3', 0, 'A', data)
        self.assertSensor('fgrid1', 49.98, 'Hz', data)
        self.assertSensor('fgrid2', -0.01, 'Hz', data)
        self.assertSensor('fgrid3', -0.01, 'Hz', data)
        self.assertSensor('pgrid1', 5955, 'W', data)
        self.assertSensor('pgrid2', 0, 'W', data)
        self.assertSensor('pgrid3', 0, 'W', data)
        self.assertSensor('total_inverter_power', 5914, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('apparent_power', 5957, 'VA', data)
        self.assertSensor('reactive_power', -6, 'var', data)
        self.assertSensor("power_factor", 0.999, "", data),
        self.assertSensor('temperature', 36.0, 'C', data)
        self.assertSensor('e_day', 4.3, 'kWh', data)
        self.assertSensor('e_total', 998.2, 'kWh', data)
        self.assertSensor('h_total', 246, 'h', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('funbit', 0, '', data)
        self.assertSensor('vbus', 397.3, 'V', data)
        self.assertSensor('vnbus', 0, 'V', data)
        self.assertSensor('derating_mode', 0, '', data)
        self.assertSensor('derating_mode_label', '', '', data)


class GW20KAU_DT_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW20KAU-DT_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW20KAU-DT_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW20KAU_DT_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW20KAU-DT', self.model_name)
        self.assertEqual('0000KDTA00000000', self.serial_number)
        self.assertEqual(15, self.dsp1_version)
        self.assertEqual(15, self.dsp2_version)
        self.assertEqual(1099, self.dsp_svn_version)
        self.assertEqual(16, self.arm_version)
        self.assertEqual(187, self.arm_svn_version)
        self.assertEqual('15.15.10', self.firmware)

    def test_GW20KAU_DT_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(42, len(data))

        self.assertSensor('timestamp', datetime.strptime('2022-10-21 19:23:42', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 390.5, 'V', data)
        self.assertSensor('ipv1', 6.8, 'A', data)
        self.assertSensor('ppv1', 2655, 'W', data)
        self.assertSensor('vpv2', 351.6, 'V', data)
        self.assertSensor('ipv2', 7.1, 'A', data)
        self.assertSensor('ppv2', 2496, 'W', data)
        self.assertSensor('ppv', 5151, 'W', data)
        self.assertSensor('vline1', 388.5, 'V', data)
        self.assertSensor('vline2', 391.7, 'V', data)
        self.assertSensor('vline3', 394.5, 'V', data)
        self.assertSensor('vgrid1', 226.1, 'V', data)
        self.assertSensor('vgrid2', 223.6, 'V', data)
        self.assertSensor('vgrid3', 228.3, 'V', data)
        self.assertSensor('igrid1', 7.2, 'A', data)
        self.assertSensor('igrid2', 7.4, 'A', data)
        self.assertSensor('igrid3', 7.1, 'A', data)
        self.assertSensor('fgrid1', 49.96, 'Hz', data)
        self.assertSensor('fgrid2', 49.96, 'Hz', data)
        self.assertSensor('fgrid3', 49.97, 'Hz', data)
        self.assertSensor('pgrid1', 1628, 'W', data)
        self.assertSensor('pgrid2', 1655, 'W', data)
        self.assertSensor('pgrid3', 1621, 'W', data)
        self.assertSensor('total_inverter_power', 4957, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor("apparent_power", 0, "VA", data),
        self.assertSensor("reactive_power", 205, "var", data),
        self.assertSensor("power_factor", 0.999, "", data),
        self.assertSensor('temperature', 36.4, 'C', data)
        self.assertSensor('e_day', 19.8, 'kWh', data)
        self.assertSensor('e_total', 4304.8, 'kWh', data)
        self.assertSensor('h_total', 1139, 'h', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('funbit', 0, '', data)
        self.assertSensor('vbus', 596.3, 'V', data)
        self.assertSensor('vnbus', 298.9, 'V', data)
        self.assertSensor('derating_mode', 4, '', data)
        self.assertSensor('derating_mode_label', 'Reactive power derating(PF/QU/FixQ)', '', data)


class GW17K_DT_Test(DtMock):

    def __init__(self, methodName='runTest'):
        DtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW17K-DT_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW17K-DT_running_data.hex')
        self.mock_response(self._READ_METER_DATA, ILLEGAL_DATA_ADDRESS)

    def test_GW20KAU_DT_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW17KT-DT', self.model_name)
        self.assertEqual('5017KDTT00BW0000', self.serial_number)
        self.assertEqual(12, self.dsp1_version)
        self.assertEqual(12, self.dsp2_version)
        self.assertEqual(931, self.dsp_svn_version)
        self.assertEqual(13, self.arm_version)
        self.assertEqual(130, self.arm_svn_version)
        self.assertEqual('12.12.0d', self.firmware)

    def test_GW20KAU_DT_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(42, len(data))

        self.assertSensor('timestamp', datetime.strptime('2024-05-20 10:35:55', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 540.0, 'V', data)
        self.assertSensor('ipv1', 10.5, 'A', data)
        self.assertSensor('ppv1', 5670, 'W', data)
        self.assertSensor('vpv2', 475.5, 'V', data)
        self.assertSensor('ipv2', 14.8, 'A', data)
        self.assertSensor('ppv2', 7037, 'W', data)
        self.assertSensor('ppv', 12707, 'W', data)
        self.assertSensor('vline1', 413.0, 'V', data)
        self.assertSensor('vline2', 411.5, 'V', data)
        self.assertSensor('vline3', 409.5, 'V', data)
        self.assertSensor('vgrid1', 236.7, 'V', data)
        self.assertSensor('vgrid2', 238.3, 'V', data)
        self.assertSensor('vgrid3', 237.3, 'V', data)
        self.assertSensor('igrid1', 17.6, 'A', data)
        self.assertSensor('igrid2', 17.5, 'A', data)
        self.assertSensor('igrid3', 17.5, 'A', data)
        self.assertSensor('fgrid1', 50.02, 'Hz', data)
        self.assertSensor('fgrid2', 50.02, 'Hz', data)
        self.assertSensor('fgrid3', 50.02, 'Hz', data)
        self.assertSensor('pgrid1', 4166, 'W', data)
        self.assertSensor('pgrid2', 4170, 'W', data)
        self.assertSensor('pgrid3', 4153, 'W', data)
        self.assertSensor('total_inverter_power', 12470, 'W', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal', '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('apparent_power', 0, 'VA', data)
        self.assertSensor('reactive_power', 0, 'var', data)
        self.assertSensor("power_factor", 0.0, "", data),
        self.assertSensor('temperature', 45.7, 'C', data)
        self.assertSensor('e_day', 29.3, 'kWh', data)
        self.assertSensor('e_total', 29984.4, 'kWh', data)
        self.assertSensor('h_total', 8357, 'h', data)
        self.assertSensor('safety_country', 1, '', data)
        self.assertSensor('safety_country_label', 'CZ-A1', '', data)
        self.assertSensor('funbit', 546, '', data)
        self.assertSensor('vbus', 621.8, 'V', data)
        self.assertSensor('vnbus', 314.2, 'V', data)
        self.assertSensor('derating_mode', 4, '', data)
        self.assertSensor('derating_mode_label', 'Reactive power derating(PF/QU/FixQ)', '', data)



from __future__ import annotations

import logging
from typing import Tuple

from .exceptions import InverterError
from .inverter import Inverter
from .inverter import OperationMode
from .inverter import SensorKind as Kind
from .protocol import ProtocolCommand, Aa55ProtocolCommand, Aa55ReadCommand, Aa55WriteCommand, Aa55WriteMultiCommand
from .sensor import *

logger = logging.getLogger(__name__)


class ES(Inverter):
    """Class representing inverter of ES/EM/BP family AKA platform 105"""

    _READ_DEVICE_VERSION_INFO: ProtocolCommand = Aa55ProtocolCommand("010200", "0182")
    _READ_DEVICE_RUNNING_DATA: ProtocolCommand = Aa55ProtocolCommand("010600", "0186")
    _READ_DEVICE_SETTINGS_DATA: ProtocolCommand = Aa55ProtocolCommand("010900", "0189")

    __sensors: Tuple[Sensor, ...] = (
        Voltage("vpv1", 0, "PV1 Voltage", Kind.PV),  # modbus 0x500
        Current("ipv1", 2, "PV1 Current", Kind.PV),
        Calculated("ppv1",
                   lambda data: round(read_voltage(data, 0) * read_current(data, 2)),
                   "PV1 Power", "W", Kind.PV),
        Byte("pv1_mode", 4, "PV1 Mode code", "", Kind.PV),
        Enum("pv1_mode_label", 4, PV_MODES, "PV1 Mode", Kind.PV),
        Voltage("vpv2", 5, "PV2 Voltage", Kind.PV),
        Current("ipv2", 7, "PV2 Current", Kind.PV),
        Calculated("ppv2",
                   lambda data: round(read_voltage(data, 5) * read_current(data, 7)),
                   "PV2 Power", "W", Kind.PV),
        Byte("pv2_mode", 9, "PV2 Mode code", "", Kind.PV),
        Enum("pv2_mode_label", 9, PV_MODES, "PV2 Mode", Kind.PV),
        Calculated("ppv",
                   lambda data: round(read_voltage(data, 0) * read_current(data, 2)) + round(
                       read_voltage(data, 5) * read_current(data, 7)),
                   "PV Power", "W", Kind.PV),
        Voltage("vbattery1", 10, "Battery Voltage", Kind.BAT),  # modbus 0x506
        # Voltage("vbattery2", 12, "Battery Voltage 2", Kind.BAT),
        Integer("battery_status", 14, "Battery Status", "", Kind.BAT),
        Temp("battery_temperature", 16, "Battery Temperature", Kind.BAT),
        Calculated("ibattery1",
                   lambda data: abs(read_current(data, 18)) * (-1 if read_byte(data, 30) == 3 else 1),
                   "Battery Current", "A", Kind.BAT),
        # round(vbattery1 * ibattery1),
        Calculated("pbattery1",
                   lambda data: abs(
                       round(read_voltage(data, 10) * read_current(data, 18))
                   ) * (-1 if read_byte(data, 30) == 3 else 1),
                   "Battery Power", "W", Kind.BAT),
        Integer("battery_charge_limit", 20, "Battery Charge Limit", "A", Kind.BAT),
        Integer("battery_discharge_limit", 22, "Battery Discharge Limit", "A", Kind.BAT),
        Integer("battery_error", 24, "Battery Error Code", "", Kind.BAT),
        Byte("battery_soc", 26, "Battery State of Charge", "%", Kind.BAT),  # modbus 0x50E
        # Byte("cbattery2", 27, "Battery State of Charge 2", "%", Kind.BAT),
        # Byte("cbattery3", 28, "Battery State of Charge 3", "%", Kind.BAT),
        Byte("battery_soh", 29, "Battery State of Health", "%", Kind.BAT),
        Byte("battery_mode", 30, "Battery Mode code", "", Kind.BAT),
        Enum("battery_mode_label", 30, BATTERY_MODES, "Battery Mode", Kind.BAT),
        Integer("battery_warning", 31, "Battery Warning", "", Kind.BAT),
        Byte("meter_status", 33, "Meter Status code", "", Kind.AC),
        Voltage("vgrid", 34, "On-grid Voltage", Kind.AC),
        Current("igrid", 36, "On-grid Current", Kind.AC),
        Calculated("pgrid",
                   lambda data: abs(read_bytes2_signed(data, 38)) * (-1 if read_byte(data, 80) == 2 else 1),
                   "On-grid Export Power", "W", Kind.AC),
        Frequency("fgrid", 40, "On-grid Frequency", Kind.AC),
        Byte("grid_mode", 42, "Work Mode code", "", Kind.GRID),
        Enum("grid_mode_label", 42, WORK_MODES_ES, "Work Mode", Kind.GRID),
        Voltage("vload", 43, "Back-up Voltage", Kind.UPS),  # modbus 0x51b
        Current("iload", 45, "Back-up Current", Kind.UPS),
        Power("pload", 47, "On-grid Power", Kind.AC),
        Frequency("fload", 49, "Back-up Frequency", Kind.UPS),
        Byte("load_mode", 51, "Load Mode code", "", Kind.AC),
        Enum("load_mode_label", 51, LOAD_MODES, "Load Mode", Kind.AC),
        Byte("work_mode", 52, "Energy Mode code", "", Kind.AC),
        Enum("work_mode_label", 52, ENERGY_MODES, "Energy Mode", Kind.AC),
        Temp("temperature", 53, "Inverter Temperature"),
        Long("error_codes", 55, "Error Codes"),
        Energy4("e_total", 59, "Total PV Generation", Kind.PV),
        Long("h_total", 63, "Hours Total", "h", Kind.PV),
        Energy("e_day", 67, "Today's PV Generation", Kind.PV),
        Energy("e_load_day", 69, "Today's Load", Kind.AC),
        Energy4("e_load_total", 71, "Total Load", Kind.AC),
        PowerS("total_power", 75, "Total Power", Kind.AC),  # modbus 0x52c
        Byte("effective_work_mode", 77, "Effective Work Mode code"),
        Integer("effective_relay_control", 78, "Effective Relay Control", "", None),
        Byte("grid_in_out", 80, "On-grid Mode code", "", Kind.GRID),
        Enum("grid_in_out_label", 80, GRID_IN_OUT_MODES, "On-grid Mode", Kind.GRID),
        Power("pback_up", 81, "Back-up Power", Kind.UPS),
        # pload + pback_up
        Calculated("plant_power",
                   lambda data: round(read_bytes2(data, 47, 0) + read_bytes2(data, 81, 0)),
                   "Plant Power", "W", Kind.AC),
        Decimal("meter_power_factor", 83, 1000, "Meter Power Factor", "", Kind.GRID),  # modbus 0x531
        # Integer("xx85", 85, "Unknown sensor@85"),
        # Integer("xx87", 87, "Unknown sensor@87"),
        Long("diagnose_result", 89, "Diag Status Code"),
        EnumBitmap4("diagnose_result_label", 89, DIAG_STATUS_CODES, "Diag Status"),
        # Energy4("e_total_exp", 93, "Total Energy (export)", Kind.GRID),
        # Energy4("e_total_imp", 97, "Total Energy (import)", Kind.GRID),
        # Voltage("vpv3", 101, "PV3 Voltage", Kind.PV),  # modbus 0x500
        # Current("ipv3", 103, "PV3 Current", Kind.PV),
        # Byte("pv3_mode", 104, "PV1 Mode", "", Kind.PV),
        # Voltage("vgrid_uo", 105, "On-grid Uo Voltage", Kind.AC),
        # Current("igrid_uo", 107, "On-grid Uo Current", Kind.AC),
        # Voltage("vgrid_wo", 109, "On-grid Wo Voltage", Kind.AC),
        # Current("igrid_wo", 111, "On-grid Wo Current", Kind.AC),
        # Energy4("e_bat_charge_total", 113, "Total Battery Charge", Kind.BAT),
        # Energy4("e_bat_discharge_total", 117, "Total Battery Discharge", Kind.BAT),

        # ppv1 + ppv2 + pbattery - pgrid
        Calculated("house_consumption",
                   lambda data:
                   round(read_voltage(data, 0) * read_current(data, 2)) +
                   round(read_voltage(data, 5) * read_current(data, 7)) +
                   (abs(round(read_voltage(data, 10) * read_current(data, 18))) *
                    (-1 if read_byte(data, 30) == 3 else 1)) -
                   (abs(read_bytes2_signed(data, 38)) * (-1 if read_byte(data, 80) == 2 else 1)),
                   "House Consumption", "W", Kind.AC),
    )

    __all_settings: Tuple[Sensor, ...] = (
        Integer("backup_supply", 12, "Backup Supply"),
        Integer("off-grid_charge", 14, "Off-grid Charge"),
        Integer("shadow_scan", 16, "Shadow Scan", "", Kind.PV),
        Integer("grid_export", 18, "Export Limit Enabled", "", Kind.GRID),
        Integer("capacity", 22, "Capacity"),
        Decimal("charge_v", 24, 10, "Charge Voltage", "V"),
        Integer("charge_i", 26, "Charge Current", "A", ),
        Integer("discharge_i", 28, "Discharge Current", "A", ),
        Decimal("discharge_v", 30, 10, "Discharge Voltage", "V"),
        Calculated("dod", lambda data: 100 - read_bytes2(data, 32, 0), "Depth of Discharge", "%"),
        Integer("battery_activated", 34, "Battery Activated"),
        Integer("bp_off_grid_charge", 36, "BP Off-grid Charge"),
        Integer("bp_pv_discharge", 38, "BP PV Discharge"),
        Integer("bp_bms_protocol", 40, "BP BMS Protocol"),
        Integer("power_factor", 42, "Power Factor"),
        Integer("grid_export_limit", 52, "Grid Export Limit", "W", Kind.GRID),
        Integer("battery_soc_protection", 56, "Battery SoC Protection", "", Kind.BAT),
        Integer("work_mode", 66, "Work Mode"),
        Integer("grid_quality_check", 68, "Grid Quality Check"),

        EcoModeV1("eco_mode_1", 1793, "Eco Mode Group 1"),  # 0x701
        ByteH("eco_mode_1_switch", 1796, "Eco Mode Group 1 Switch", "", Kind.BAT),
        EcoModeV1("eco_mode_2", 1797, "Eco Mode Group 2"),
        ByteH("eco_mode_2_switch", 1800, "Eco Mode Group 2 Switch", "", Kind.BAT),
        EcoModeV1("eco_mode_3", 1801, "Eco Mode Group 3"),
        ByteH("eco_mode_3_switch", 1804, "Eco Mode Group 3 Switch", "", Kind.BAT),
        EcoModeV1("eco_mode_4", 1805, "Eco Mode Group 4"),
        ByteH("eco_mode_4_switch", 1808, "Eco Mode Group 4 Switch", "", Kind.BAT),
    )

    # Settings added in ARM firmware 14
    __settings_arm_fw_14: Tuple[Sensor, ...] = (
        EcoModeV2("eco_mode_1", 47547, "Eco Mode Group 1"),
        ByteH("eco_mode_1_switch", 47549, "Eco Mode Group 1 Switch"),
        EcoModeV2("eco_mode_2", 47553, "Eco Mode Group 2"),
        ByteH("eco_mode_2_switch", 47555, "Eco Mode Group 2 Switch"),
        EcoModeV2("eco_mode_3", 47559, "Eco Mode Group 3"),
        ByteH("eco_mode_3_switch", 47561, "Eco Mode Group 3 Switch"),
        EcoModeV2("eco_mode_4", 47565, "Eco Mode Group 4"),
        ByteH("eco_mode_4_switch", 47567, "Eco Mode Group 4 Switch"),
    )

    def __init__(self, host: str, port: int, comm_addr: int = 0, timeout: int = 1, retries: int = 3):
        super().__init__(host, port, comm_addr if comm_addr else 0xf7, timeout, retries)
        self._settings: dict[str, Sensor] = {s.id_: s for s in self.__all_settings}

    def _supports_eco_mode_v2(self) -> bool:
        if self.arm_version < 14:
            return False
        if "EMU" in self.serial_number or "EMJ" in self.serial_number:
            return self.dsp1_version >= 11
        if "ESU" in self.serial_number or "ESA" in self.serial_number:
            return self.dsp1_version >= 22
        if "BPS" in self.serial_number or "BPU" in self.serial_number:
            return self.dsp1_version >= 10
        return False

    async def read_device_info(self):
        response = await self._read_from_socket(self._READ_DEVICE_VERSION_INFO)
        response = response.response_data()
        self.firmware = self._decode(response[0:5]).rstrip()
        self.model_name = self._decode(response[5:15]).rstrip()
        self.serial_number = self._decode(response[31:47])
        self.arm_firmware = self._decode(response[51:63])  # AKA software_version
        try:
            if len(self.firmware) >= 2:
                self.dsp1_version = int(self.firmware[0:2])
            if len(self.firmware) >= 4:
                self.dsp2_version = int(self.firmware[2:4])
            if len(self.firmware) >= 5:
                self.arm_version = int(self.firmware[4], base=36)
        except ValueError:
            logger.exception("Error decoding firmware version %s.", self.firmware)

        if self._supports_eco_mode_v2():
            self._settings.update({s.id_: s for s in self.__settings_arm_fw_14})

    async def read_runtime_data(self) -> Dict[str, Any]:
        response = await self._read_from_socket(self._READ_DEVICE_RUNNING_DATA)
        data = self._map_response(response, self.__sensors)
        return data

    async def read_setting(self, setting_id: str) -> Any:
        if setting_id == 'time':
            # Fake setting, just to enable write_setting to work (if checked as pair in read as in HA)
            # There does not seem to be time setting/sensor available (or is not known)
            return datetime.now()
        elif setting_id in ('eco_mode_1', 'eco_mode_2', 'eco_mode_3', 'eco_mode_4'):
            setting: Sensor | None = self._settings.get(setting_id)
            if not setting:
                raise ValueError(f'Unknown setting "{setting_id}"')
            return await self._read_setting(setting)
        elif setting_id.startswith("modbus"):
            response = await self._read_from_socket(self._read_command(int(setting_id[7:]), 1))
            return int.from_bytes(response.read(2), byteorder="big", signed=True)
        elif setting_id in self._settings:
            logger.debug("Reading setting %s", setting_id)
            all_settings = await self.read_settings_data()
            return all_settings.get(setting_id)
        else:
            raise ValueError(f'Unknown setting "{setting_id}"')

    async def _read_setting(self, setting: Sensor) -> Any:
        count = (setting.size_ + (setting.size_ % 2)) // 2
        if self._is_modbus_setting(setting):
            response = await self._read_from_socket(self._read_command(setting.offset, count))
            return setting.read_value(response)
        else:
            response = await self._read_from_socket(Aa55ReadCommand(setting.offset, count))
            return setting.read_value(response)

    async def write_setting(self, setting_id: str, value: Any):
        if setting_id == 'time':
            await self._read_from_socket(
                Aa55ProtocolCommand("030206" + Timestamp("time", 0, "").encode_value(value).hex(), "0382")
            )
        elif setting_id.startswith("modbus"):
            await self._read_from_socket(self._write_command(int(setting_id[7:]), int(value)))
        else:
            setting: Sensor | None = self._settings.get(setting_id)
            if not setting:
                raise ValueError(f'Unknown setting "{setting_id}"')
            await self._write_setting(setting, value)

    async def _write_setting(self, setting: Sensor, value: Any):
        if setting.size_ == 1:
            # modbus can address/store only 16 bit values, read the other 8 bytes
            if self._is_modbus_setting(setting):
                response = await self._read_from_socket(self._read_command(setting.offset, 1))
            else:
                response = await self._read_from_socket(Aa55ReadCommand(setting.offset, 1))
            raw_value = setting.encode_value(value, response.response_data()[0:2])
        else:
            raw_value = setting.encode_value(value)
        if len(raw_value) <= 2:
            value = int.from_bytes(raw_value, byteorder="big", signed=True)
            if self._is_modbus_setting(setting):
                await self._read_from_socket(self._write_command(setting.offset, value))
            else:
                await self._read_from_socket(Aa55WriteCommand(setting.offset, value))
        else:
            if self._is_modbus_setting(setting):
                await self._read_from_socket(self._write_multi_command(setting.offset, raw_value))
            else:
                await self._read_from_socket(Aa55WriteMultiCommand(setting.offset, raw_value))

    async def read_settings_data(self) -> Dict[str, Any]:
        response = await self._read_from_socket(self._READ_DEVICE_SETTINGS_DATA)
        data = self._map_response(response, self.settings())
        return data

    async def get_grid_export_limit(self) -> int:
        return await self.read_setting('grid_export_limit')

    async def set_grid_export_limit(self, export_limit: int) -> None:
        if export_limit >= 0:
            await self._read_from_socket(
                Aa55ProtocolCommand("033502" + "{:04x}".format(export_limit), "03b5")
            )

    async def get_operation_modes(self, include_emulated: bool) -> Tuple[OperationMode, ...]:
        result = [e for e in OperationMode]
        result.remove(OperationMode.PEAK_SHAVING)
        result.remove(OperationMode.SELF_USE)
        if not include_emulated:
            result.remove(OperationMode.ECO_CHARGE)
            result.remove(OperationMode.ECO_DISCHARGE)
        return tuple(result)

    async def get_operation_mode(self) -> OperationMode | None:
        mode_id = await self.read_setting('work_mode')
        try:
            mode = OperationMode(mode_id)
        except ValueError:
            logger.debug("Unknown work_mode value %s", mode_id)
            return None
        if OperationMode.ECO != mode:
            return mode
        eco_mode = await self.read_setting('eco_mode_1')
        if eco_mode.is_eco_charge_mode():
            return OperationMode.ECO_CHARGE
        elif eco_mode.is_eco_discharge_mode():
            return OperationMode.ECO_DISCHARGE
        else:
            return OperationMode.ECO

    async def set_operation_mode(self, operation_mode: OperationMode, eco_mode_power: int = 100,
                                 eco_mode_soc: int = 100) -> None:
        if operation_mode == OperationMode.GENERAL:
            await self._set_general_mode()
        elif operation_mode == OperationMode.OFF_GRID:
            await self._set_offgrid_mode()
        elif operation_mode == OperationMode.BACKUP:
            await self._set_backup_mode()
        elif operation_mode == OperationMode.ECO:
            await self._set_eco_mode()
        elif operation_mode == OperationMode.PEAK_SHAVING:
            raise InverterError("Operation not supported.")
        elif operation_mode in (OperationMode.ECO_CHARGE, OperationMode.ECO_DISCHARGE):
            if eco_mode_power < 0 or eco_mode_power > 100:
                raise ValueError()
            if eco_mode_soc < 0 or eco_mode_soc > 100:
                raise ValueError()
            eco_mode: EcoMode | Sensor = self._settings.get('eco_mode_1')
            await self._read_setting(eco_mode)
            if operation_mode == OperationMode.ECO_CHARGE:
                await self.write_setting('eco_mode_1', eco_mode.encode_charge(eco_mode_power, eco_mode_soc))
            else:
                await self.write_setting('eco_mode_1', eco_mode.encode_discharge(eco_mode_power))
            await self.write_setting('eco_mode_2_switch', 0)
            await self.write_setting('eco_mode_3_switch', 0)
            await self.write_setting('eco_mode_4_switch', 0)
            await self._set_eco_mode()

    async def get_ongrid_battery_dod(self) -> int:
        return await self.read_setting('dod')

    async def set_ongrid_battery_dod(self, dod: int) -> None:
        if 0 <= dod <= 100:
            await self._read_from_socket(Aa55WriteCommand(0x560, 100 - dod))

    async def _reset_inverter(self) -> None:
        await self._read_from_socket(Aa55ProtocolCommand("031d00", "039d"))

    def sensors(self) -> Tuple[Sensor, ...]:
        return self.__sensors

    def settings(self) -> Tuple[Sensor, ...]:
        return tuple(self._settings.values())

    async def _set_general_mode(self) -> None:
        if self.arm_version >= 7:
            if self._supports_eco_mode_v2():
                await self._clear_battery_mode_param()
            else:
                await self._set_limit_power_for_charge(0, 0, 0, 0, 0)
                await self._set_limit_power_for_discharge(0, 0, 0, 0, 0)
                await self._clear_battery_mode_param()
        else:
            await self._set_limit_power_for_charge(0, 0, 0, 0, 0)
            await self._set_limit_power_for_discharge(0, 0, 0, 0, 0)
        await self._set_offgrid_work_mode(0)
        await self._set_work_mode(OperationMode.GENERAL)

    async def _set_offgrid_mode(self) -> None:
        if self.arm_version >= 7:
            await self._clear_battery_mode_param()
        else:
            await self._set_limit_power_for_charge(0, 0, 23, 59, 0)
            await self._set_limit_power_for_discharge(0, 0, 0, 0, 0)
        await self._set_offgrid_work_mode(1)
        await self._set_relay_control(3)
        await self._set_store_energy_mode(0)
        await self._set_work_mode(OperationMode.OFF_GRID)

    async def _set_backup_mode(self) -> None:
        if self.arm_version >= 7:
            if self._supports_eco_mode_v2():
                await self._clear_battery_mode_param()
            else:
                await self._clear_battery_mode_param()
                await self._set_limit_power_for_charge(0, 0, 23, 59, 10)
        else:
            await self._set_limit_power_for_charge(0, 0, 23, 59, 10)
            await self._set_limit_power_for_discharge(0, 0, 0, 0, 0)
        await self._set_offgrid_work_mode(0)
        await self._set_work_mode(OperationMode.BACKUP)

    async def _set_eco_mode(self) -> None:
        await self._set_offgrid_work_mode(0)
        await self._set_work_mode(OperationMode.ECO)

    async def _clear_battery_mode_param(self) -> None:
        await self._read_from_socket(Aa55WriteCommand(0x0700, 1))

    async def _set_limit_power_for_charge(self, startH: int, startM: int, stopH: int, stopM: int, limit: int) -> None:
        if limit < 0 or limit > 100:
            raise ValueError()
        await self._read_from_socket(Aa55ProtocolCommand("032c05"
                                                         + "{:02x}".format(startH) + "{:02x}".format(startM)
                                                         + "{:02x}".format(stopH) + "{:02x}".format(stopM)
                                                         + "{:02x}".format(limit), "03AC"))

    async def _set_limit_power_for_discharge(self, startH: int, startM: int, stopH: int, stopM: int,
                                             limit: int) -> None:
        if limit < 0 or limit > 100:
            raise ValueError()
        await self._read_from_socket(Aa55ProtocolCommand("032d05"
                                                         + "{:02x}".format(startH) + "{:02x}".format(startM)
                                                         + "{:02x}".format(stopH) + "{:02x}".format(stopM)
                                                         + "{:02x}".format(limit), "03AD"))

    async def _set_offgrid_work_mode(self, mode: int) -> None:
        await self._read_from_socket(Aa55ProtocolCommand("033601" + "{:02x}".format(mode), "03B6"))

    async def _set_relay_control(self, mode: int) -> None:
        param = 0
        if mode == 2:
            param = 16
        elif mode == 3:
            param = 48
        await self._read_from_socket(Aa55ProtocolCommand("03270200" + "{:02x}".format(param), "03B7"))

    async def _set_store_energy_mode(self, mode: int) -> None:
        param = 0
        if mode == 0:
            param = 4
        elif mode == 1:
            param = 2
        elif mode == 2:
            param = 8
        elif mode == 3:
            param = 1
        await self._read_from_socket(Aa55ProtocolCommand("032601" + "{:02x}".format(param), "03B6"))

    async def _set_work_mode(self, mode: int) -> None:
        await self._read_from_socket(Aa55ProtocolCommand("035901" + "{:02x}".format(mode), "03D9"))

    def _is_modbus_setting(self, sensor: Sensor) -> bool:
        return sensor.offset > 30000



"""Simple test script to check inverter UDP protocol communication"""
import asyncio
import logging
import sys
from importlib.metadata import version

# Force the local files, not pip installed lib
sys.path.insert(0, '../goodwe')
sys.path.insert(0, '../../goodwe')

import goodwe

logging.basicConfig(
    format="%(asctime)-15s %(funcName)s(%(lineno)d) - %(levelname)s: %(message)s",
    stream=sys.stderr,
    level=getattr(logging, "DEBUG", None),
)

if sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

module_ver = None
try:
    module_ver = version('goodwe')
except ModuleNotFoundError:
    pass

if module_ver:
    print("WARNING !!!")
    print("==============================")
    print(f"You are executing code with installed pip version goodwe:{module_ver}")
    print("You are not testing the local files, if that was what you meant !!!")
    print("==============================")

# Set the appropriate IP address
IP_ADDRESS = "192.168.2.14"
PORT = 8899
FAMILY = "ET"  # One of ET, EH, ES, EM, DT, NS, XS or None to detect family automatically
COMM_ADDR = 0xf7  # Usually 0xf7 for ET/EH/EM/ES or 0x7f for DT/D-NS/XS, or None for default value
TIMEOUT = 1
RETRIES = 3

inverter = asyncio.run(goodwe.connect(IP_ADDRESS, PORT, FAMILY, COMM_ADDR, TIMEOUT, RETRIES))
print(f"Identified inverter\n"
      f"- Model: {inverter.model_name}\n"
      f"- SerialNr: {inverter.serial_number}\n"
      f"- Rated power: {inverter.rated_power}\n"
      f"- A/C output type: {inverter.ac_output_type}\n"
      f"- Firmware: {inverter.firmware}\n"
      f"- ARM firmware: {inverter.arm_firmware}\n"
      f"- Modbus version: {inverter.modbus_version}\n"
      f"- DSP1 version: {inverter.dsp1_version}\n"
      f"- DSP2 version: {inverter.dsp2_version}\n"
      f"- DSP svn version: {inverter.dsp_svn_version}\n"
      f"- Arm version: {inverter.arm_version}\n"
      f"- ARM svn version: {inverter.arm_svn_version}\n"
      )

# -----------------
# Read runtime data
# -----------------
response = asyncio.run(inverter.read_runtime_data())
for sensor in inverter.sensors():
    if sensor.id_ in response:
        print(f"{sensor.id_}: \t\t {sensor.name} = {response[sensor.id_]} {sensor.unit}")

# -------------
# Read settings
# -------------
# response = asyncio.run(inverter.read_settings_data())
# for setting in inverter.settings():
#    if setting.id_ in response:
#        print(f"{setting.id_}: \t\t {setting.name} = {response[setting.id_]} {setting.unit}")

# -----------------
# Set inverter time
# -----------------
# print(asyncio.run(inverter.read_setting('time')))
# asyncio.run(inverter.write_setting('time', datetime.datetime.now()))
# print(asyncio.run(inverter.read_setting('time')))

# ------------------------------
# Set inverter grid export limit
# ------------------------------
# print(asyncio.run(inverter.get_grid_export_limit()))
# asyncio.run(inverter.set_grid_export_limit(4000))
# print(asyncio.run(inverter.get_grid_export_limit()))

# ---------------------------
# Set inverter operation mode
# ---------------------------
# print(asyncio.run(inverter.get_operation_mode()))
# asyncio.run(inverter.set_operation_mode(goodwe.OperationMode.BACKUP))
# print(asyncio.run(inverter.get_operation_mode()))

# --------------------
# Set inverter setting
# --------------------
# print(asyncio.run(inverter.read_setting('grid_export_limit')))
# asyncio.run(inverter.write_setting('grid_export_limit', 4000))
# print(asyncio.run(inverter.read_setting('grid_export_limit')))

# --------------------
# Get inverter modbus setting
# --------------------
# print(asyncio.run(inverter.read_setting('modbus-47000')))

# -------------------------------
# Execute modbus RTU protocol command
# -------------------------------
# response = asyncio.run(goodwe.protocol.ModbusRtuReadCommand(COMM_ADDR, 0x88b8, 0x21).execute(
#    goodwe.protocol.UdpInverterProtocol(IP_ADDRESS, PORT, TIMEOUT, RETRIES)))
# print(response)

# -------------------------------
# Execute modbus TCP protocol command
# -------------------------------
# response = asyncio.run(goodwe.protocol.ModbusTcpReadCommand(180, 301, 3).execute(
#    goodwe.protocol.TcpInverterProtocol('192.168.1.13', 502, TIMEOUT, RETRIES)))
# print(response.response_data().hex())

# -------------------------------
# Execute AA55 protocol command
# -------------------------------
# response = asyncio.run(goodwe.protocol.Aa55ProtocolCommand("010200", "0182").execute(
#    goodwe.protocol.UdpInverterProtocol(IP_ADDRESS, PORT, TIMEOUT, RETRIES)))
# print(response)

# -----------------
# Test parallel requests
#
# async def run_in_parallel(inverter):
#    a, b, c, = await asyncio.gather(inverter.get_grid_export_limit(), inverter.get_ongrid_battery_dod(),
#                                    inverter.read_runtime_data())
#    print(a)
#    print(b)
#    print(c)
#
# asyncio.run(run_in_parallel(inverter))



import asyncio
import os
from datetime import datetime
from unittest import TestCase

from goodwe.et import ET
from goodwe.exceptions import RequestRejectedException, RequestFailedException
from goodwe.inverter import OperationMode
from goodwe.modbus import ILLEGAL_DATA_ADDRESS
from goodwe.protocol import ModbusRtuReadCommand, ProtocolCommand, ProtocolResponse


class EtMock(TestCase, ET):

    def __init__(self, methodName='runTest'):
        TestCase.__init__(self, methodName)
        ET.__init__(self, "localhost", 8899)
        self.sensor_map = {s.id_: s for s in self.sensors()}
        self._mock_responses = {}
        self._list_of_requests = []

    def mock_response(self, command: ProtocolCommand, filename: str):
        self._mock_responses[command] = filename

    async def _read_from_socket(self, command: ProtocolCommand) -> ProtocolResponse:
        """Mock UDP communication"""
        root_dir = os.path.dirname(os.path.abspath(__file__))
        filename = self._mock_responses.get(command)
        if filename is not None:
            if ILLEGAL_DATA_ADDRESS == filename:
                raise RequestRejectedException(ILLEGAL_DATA_ADDRESS)
            if 'NO RESPONSE' == filename:
                raise RequestFailedException()
            with open(root_dir + '/sample/et/' + filename, 'r') as f:
                response = bytes.fromhex(f.read())
                if not command.validator(response):
                    raise RequestFailedException
                return ProtocolResponse(response, command)
        else:
            self.request = command.request
            self._list_of_requests.append(command.request)
            return ProtocolResponse(bytes.fromhex("aa55f700010203040506070809"), command)

    def assertSensor(self, sensor_name, expected_value, expected_unit, data):
        self.assertEqual(expected_value, data.get(sensor_name))
        sensor = self.sensor_map.get(sensor_name);
        self.assertEqual(expected_unit, sensor.unit)
        self.sensor_map.pop(sensor_name)

    @classmethod
    def setUpClass(cls):
        cls.loop = asyncio.get_event_loop()


class GW10K_ET_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW10K-ET_device_info_fw617.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW10K-ET_running_data.hex')
        self.mock_response(self._READ_METER_DATA, 'GW10K-ET_meter_data.hex')
        self.mock_response(self._READ_BATTERY_INFO, 'GW10K-ET_battery_info.hex')
        self.mock_response(ModbusRtuReadCommand(0xf7, 47547, 6), ILLEGAL_DATA_ADDRESS)
        self.mock_response(ModbusRtuReadCommand(0xf7, 47589, 6), ILLEGAL_DATA_ADDRESS)
        self.mock_response(ModbusRtuReadCommand(0xf7, 47515, 4), 'eco_mode_v1.hex')

    def test_GW10K_ET_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW10K-ET', self.model_name)
        self.assertEqual('9010KETU000W0000', self.serial_number)
        self.assertEqual(10000, self.rated_power)
        self.assertEqual(1, self.modbus_version)
        self.assertEqual(254, self.ac_output_type)
        self.assertEqual(6, self.dsp1_version)
        self.assertEqual(6, self.dsp2_version)
        self.assertEqual(152, self.dsp_svn_version)
        self.assertEqual(17, self.arm_version)
        self.assertEqual(192, self.arm_svn_version)
        self.assertEqual('04029-06-S11', self.firmware)
        self.assertEqual('02041-17-S00', self.arm_firmware)

    def test_GW10K_ET_runtime_data(self):
        # Reset sensors
        self.loop.run_until_complete(self.read_device_info())
        self.sensor_map = {s.id_: s for s in self.sensors()}

        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(145, len(data))

        self.assertEqual(36015, self.sensor_map.get("meter_e_total_exp").offset)

        # for sensor in self.sensors():
        #   print(f"self.assertSensor('{sensor.id_}', {data[sensor.id_]}, '{self.sensor_map.get(sensor.id_).unit}', data)")

        self.assertSensor('timestamp', datetime.strptime('2021-08-22 11:11:12', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 332.6, 'V', data)
        self.assertSensor('ipv1', 5.1, 'A', data)
        self.assertSensor('ppv1', 1695, 'W', data)
        self.assertSensor('vpv2', 332.6, 'V', data)
        self.assertSensor('ipv2', 5.3, 'A', data)
        self.assertSensor('ppv2', 1761, 'W', data)
        self.assertSensor('ppv', 3456, 'W', data)
        self.assertSensor('pv1_mode', 2, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv2_mode', 2, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('vgrid', 239.3, 'V', data)
        self.assertSensor('igrid', 1.5, 'A', data)
        self.assertSensor('fgrid', 49.99, 'Hz', data)
        self.assertSensor('pgrid', 336, 'W', data)
        self.assertSensor('vgrid2', 241.5, 'V', data)
        self.assertSensor('igrid2', 1.3, 'A', data)
        self.assertSensor('fgrid2', 49.99, 'Hz', data)
        self.assertSensor('pgrid2', 287, 'W', data)
        self.assertSensor('vgrid3', 241.1, 'V', data)
        self.assertSensor('igrid3', 1.1, 'A', data)
        self.assertSensor('fgrid3', 49.99, 'Hz', data)
        self.assertSensor('pgrid3', 206, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 831, 'W', data)
        self.assertSensor('active_power', -3, 'W', data)
        self.assertSensor('grid_in_out', 0, '', data)
        self.assertSensor('grid_in_out_label', 'Idle', '', data)
        self.assertSensor('reactive_power', 0, 'var', data)
        self.assertSensor('apparent_power', 0, 'VA', data)
        self.assertSensor('backup_v1', 239.0, 'V', data)
        self.assertSensor('backup_i1', 0.6, 'A', data)
        self.assertSensor('backup_f1', 49.98, 'Hz', data)
        self.assertSensor('load_mode1', 1, '', data)
        self.assertSensor('backup_p1', 107, 'W', data)
        self.assertSensor('backup_v2', 241.3, 'V', data)
        self.assertSensor('backup_i2', 0.9, 'A', data)
        self.assertSensor('backup_f2', 50.0, 'Hz', data)
        self.assertSensor('load_mode2', 1, '', data)
        self.assertSensor('backup_p2', 189, 'W', data)
        self.assertSensor('backup_v3', 241.2, 'V', data)
        self.assertSensor('backup_i3', 0.2, 'A', data)
        self.assertSensor('backup_f3', 49.99, 'Hz', data)
        self.assertSensor('load_mode3', 1, '', data)
        self.assertSensor('backup_p3', 0, 'W', data)
        self.assertSensor('load_p1', 224, 'W', data)
        self.assertSensor('load_p2', 80, 'W', data)
        self.assertSensor('load_p3', 233, 'W', data)
        self.assertSensor('load_ptotal', 522, 'W', data)
        self.assertSensor('backup_ptotal', 312, 'W', data)
        self.assertSensor('ups_load', 4, '%', data)
        self.assertSensor('temperature_air', 51.0, 'C', data)
        self.assertSensor('temperature_module', 0, 'C', data)
        self.assertSensor('temperature', 58.7, 'C', data)
        self.assertSensor('function_bit', 0, '', data)
        self.assertSensor('bus_voltage', 803.6, 'V', data)
        self.assertSensor('nbus_voltage', 401.8, 'V', data)
        self.assertSensor('vbattery1', 254.2, 'V', data)
        self.assertSensor('ibattery1', -9.8, 'A', data)
        self.assertSensor('pbattery1', -2512, 'W', data)
        self.assertSensor('battery_mode', 3, '', data)
        self.assertSensor('battery_mode_label', 'Charge', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor("e_total", 6085.3, 'kWh', data)
        self.assertSensor("e_day", 12.5, 'kWh', data)
        self.assertSensor("e_total_exp", 4718.6, 'kWh', data)
        self.assertSensor('h_total', 9246, 'h', data)
        self.assertSensor("e_day_exp", 9.8, 'kWh', data)
        self.assertSensor("e_total_imp", 58.0, 'kWh', data)
        self.assertSensor("e_day_imp", 0, 'kWh', data)
        self.assertSensor("e_load_total", 8820.2, 'kWh', data)
        self.assertSensor("e_load_day", 11.6, 'kWh', data)
        self.assertSensor("e_bat_charge_total", 2758.1, 'kWh', data)
        self.assertSensor("e_bat_charge_day", 5.3, 'kWh', data)
        self.assertSensor("e_bat_discharge_total", 2442.1, 'kWh', data)
        self.assertSensor("e_bat_discharge_day", 2.9, 'kWh', data)
        self.assertSensor('diagnose_result', 117442560, '', data)
        self.assertSensor('diagnose_result_label',
                          'Self-use load light, Export power limit set, PF value set, Real power limit set',
                          '', data)
        self.assertSensor('house_consumption', 947, 'W', data)
        self.assertSensor('battery_bms', 255, '', data)
        self.assertSensor('battery_index', 256, '', data)
        self.assertSensor('battery_status', 1, '', data)
        self.assertSensor('battery_temperature', 35.0, 'C', data)
        self.assertSensor('battery_charge_limit', 25, 'A', data)
        self.assertSensor('battery_discharge_limit', 25, 'A', data)
        self.assertSensor("battery_error_l", 0, "", data),
        self.assertSensor('battery_soc', 68, '%', data)
        self.assertSensor('battery_soh', 99, '%', data)
        self.assertSensor("battery_modules", 5, "", data)
        self.assertSensor("battery_warning_l", 0, "", data)
        self.assertSensor("battery_protocol", 257, "", data)
        self.assertSensor("battery_error_h", 0, "", data)
        self.assertSensor("battery_error", "", "", data)
        self.assertSensor("battery_warning_h", 0, "", data)
        self.assertSensor("battery_warning", "", "", data)
        self.assertSensor("battery_sw_version", 0, "", data)
        self.assertSensor("battery_hw_version", 0, "", data)
        self.assertSensor("battery_max_cell_temp_id", 0, "", data)
        self.assertSensor("battery_min_cell_temp_id", 0, "", data)
        self.assertSensor("battery_max_cell_voltage_id", 0, "", data)
        self.assertSensor("battery_min_cell_voltage_id", 0, "", data)
        self.assertSensor("battery_max_cell_temp", 0, "C", data)
        self.assertSensor("battery_min_cell_temp", 0, "C", data)
        self.assertSensor("battery_max_cell_voltage", 0, "V", data)
        self.assertSensor("battery_min_cell_voltage", 0, "V", data)
        self.assertSensor('commode', 1, '', data)
        self.assertSensor('rssi', 35, '', data)
        self.assertSensor('manufacture_code', 10, '', data)
        self.assertSensor('meter_test_status', 0, '', data)
        self.assertSensor('meter_comm_status', 1, '', data)
        self.assertSensor('active_power1', -57, 'W', data)
        self.assertSensor('active_power2', -46, 'W', data)
        self.assertSensor('active_power3', -6, 'W', data)
        self.assertSensor('active_power_total', -110, 'W', data)
        self.assertSensor('reactive_power_total', 1336, 'var', data)
        self.assertSensor('meter_power_factor1', -0.145, '', data)
        self.assertSensor('meter_power_factor2', -0.124, '', data)
        self.assertSensor('meter_power_factor3', -0.014, '', data)
        self.assertSensor('meter_power_factor', -0.08, '', data)
        self.assertSensor('meter_freq', 50.05, 'Hz', data)
        self.assertSensor('meter_e_total_exp', 10.514, 'kWh', data)
        self.assertSensor('meter_e_total_imp', 3254.462, 'kWh', data)
        self.assertSensor('meter_active_power1', -57, 'W', data)
        self.assertSensor('meter_active_power2', -46, 'W', data)
        self.assertSensor('meter_active_power3', -6, 'W', data)
        self.assertSensor('meter_active_power_total', -110, 'W', data)
        self.assertSensor('meter_reactive_power1', 364, 'var', data)
        self.assertSensor('meter_reactive_power2', 357, 'var', data)
        self.assertSensor('meter_reactive_power3', 614, 'var', data)
        self.assertSensor('meter_reactive_power_total', 1336, 'var', data)
        self.assertSensor('meter_apparent_power1', -402, 'VA', data)
        self.assertSensor('meter_apparent_power2', -372, 'VA', data)
        self.assertSensor('meter_apparent_power3', -627, 'VA', data)
        self.assertSensor('meter_apparent_power_total', -1403, 'VA', data)
        self.assertSensor('meter_type', 1, '', data)
        self.assertSensor('meter_sw_version', 3, '', data)

        self.assertFalse(self.sensor_map, f"Some sensors were not tested {self.sensor_map}")

    def test_GW10K_ET_setting(self):
        self.assertEqual(66, len(self.settings()))
        settings = {s.id_: s for s in self.settings()}
        self.assertEqual('Timestamp', type(settings.get("time")).__name__)
        self.assertEqual('EcoModeV1', type(settings.get("eco_mode_1")).__name__)

    def test_GW10K_ET_read_setting(self):
        self.loop.run_until_complete(self.read_setting('work_mode'))
        self.assertEqual('f703b798000136c7', self.request.hex())

        self.loop.run_until_complete(self.read_setting('grid_export_limit'))
        self.assertEqual('f703b996000155ec', self.request.hex())

        self.loop.run_until_complete(self.read_setting('time'))
        self.assertEqual('f703b090000337b0', self.request.hex())

        self.loop.run_until_complete(self.read_setting('modbus_47000'))
        self.assertEqual('f703b798000136c7', self.request.hex())

    def test_GW10K_ET_write_setting(self):
        self.loop.run_until_complete(self.write_setting('grid_export_limit', 100))
        self.assertEqual('f706b996006459c7', self.request.hex())

        self.loop.run_until_complete(self.write_setting('modbus_47510', 100))
        self.assertEqual('f706b996006459c7', self.request.hex())

        self.loop.run_until_complete(self.write_setting('time', datetime(2022, 1, 4, 18, 30, 25)))
        self.assertEqual('f710b090000306160104121e19a961', self.request.hex())

    def test_get_grid_export_limit(self):
        self.loop.run_until_complete(self.get_grid_export_limit())
        self.assertEqual('f703b996000155ec', self.request.hex())

    def test_set_grid_export_limit(self):
        self.loop.run_until_complete(self.set_grid_export_limit(100))
        self.assertEqual('f706b996006459c7', self.request.hex())

    def test_get_operation_modes(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual((OperationMode.GENERAL, OperationMode.OFF_GRID, OperationMode.BACKUP, OperationMode.ECO),
                         self.loop.run_until_complete(self.get_operation_modes(False)))
        self.assertEqual((OperationMode.GENERAL, OperationMode.OFF_GRID, OperationMode.BACKUP, OperationMode.ECO,
                          OperationMode.ECO_CHARGE, OperationMode.ECO_DISCHARGE),
                         self.loop.run_until_complete(self.get_operation_modes(True)))

    # def test_get_operation_mode(self):
    #    self.loop.run_until_complete(self.get_operation_mode())
    #    self.assertEqual('f703b798000136c7', self.request.hex())

    #    def test_set_operation_mode(self):
    #        self.loop.run_until_complete(self.set_operation_mode(1))
    #        self.assertEqual('f706b7980001fac7', self.request.hex())

    def test_set_operation_mode_ECO_CHARGE(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.set_operation_mode(OperationMode.ECO_CHARGE, eco_mode_power=40))
        self.assertEqual('f710b99b0004080000173bffd8ff7f1343', self._list_of_requests[-9].hex())
        self.loop.run_until_complete(
            self.set_operation_mode(OperationMode.ECO_CHARGE, eco_mode_power=40, eco_mode_soc=80))
        self.assertEqual('f710b99b0004080000173bffd8ff7f1343', self._list_of_requests[-9].hex())

    def test_set_operation_mode_DISCHARGE(self):
        self.loop.run_until_complete(self.read_device_info())
        self.loop.run_until_complete(self.set_operation_mode(OperationMode.ECO_DISCHARGE, eco_mode_power=50))
        self.assertEqual('f710b99b0004080000173b0032ff7f02a3', self._list_of_requests[-9].hex())

    def test_get_ongrid_battery_dod(self):
        self.loop.run_until_complete(self.get_ongrid_battery_dod())
        self.assertEqual('f703b12c00017669', self.request.hex())

    def test_set_ongrid_battery_dod(self):
        self.loop.run_until_complete(self.set_ongrid_battery_dod(80))
        self.assertEqual('f706b12c00147ba6', self.request.hex())


class GW10K_ET_fw819_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW10K-ET_device_info_fw819.hex')
        self.mock_response(ModbusRtuReadCommand(0xf7, 47547, 6), 'eco_mode_v2.hex')
        self.mock_response(ModbusRtuReadCommand(0xf7, 47589, 6), ILLEGAL_DATA_ADDRESS)
        asyncio.get_event_loop().run_until_complete(self.read_device_info())

    def test_GW10K_ET_fw819_device_info(self):
        self.assertEqual('0GW10K-ET', self.model_name)
        self.assertEqual('9010KETU00000000', self.serial_number)
        self.assertEqual(10000, self.rated_power)
        self.assertEqual(1, self.modbus_version)
        self.assertEqual(254, self.ac_output_type)
        self.assertEqual(8, self.dsp1_version)
        self.assertEqual(8, self.dsp2_version)
        self.assertEqual(159, self.dsp_svn_version)
        self.assertEqual(19, self.arm_version)
        self.assertEqual(207, self.arm_svn_version)
        self.assertEqual('04029-08-S11', self.firmware)
        self.assertEqual('02041-19-S00', self.arm_firmware)

    def test_GW10K_ET_settings_fw819(self):
        self.assertEqual(73, len(self.settings()))
        settings = {s.id_: s for s in self.settings()}
        self.assertEqual('EcoModeV2', type(settings.get("eco_mode_1")).__name__)
        self.assertEqual(None, settings.get("peak_shaving_mode"))

    def test_set_operation_mode_ECO_CHARGE(self):
        self.loop.run_until_complete(
            self.set_operation_mode(OperationMode.ECO_CHARGE, eco_mode_power=40, eco_mode_soc=80))
        self.assertEqual('f710b9bb00060c0000173bff7fffd80050000002cc', self._list_of_requests[-9].hex())
        self.loop.run_until_complete(
            self.set_operation_mode(OperationMode.ECO_CHARGE, eco_mode_power=40))
        self.assertEqual('f710b9bb00060c0000173bff7fffd8006400004302', self._list_of_requests[-9].hex())

    def test_set_operation_mode_ECO_DISCHARGE(self):
        self.loop.run_until_complete(self.set_operation_mode(OperationMode.ECO_DISCHARGE, eco_mode_power=50))
        self.assertEqual('f710b9bb00060c0000173bff7f0032006400004eda', self._list_of_requests[-9].hex())


class GW10K_ET_fw1023_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW10K-ET_device_info_fw1023.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW10K-ET_running_data_fw1023.hex')
        asyncio.get_event_loop().run_until_complete(self.read_device_info())

    def test_GW10K_ET_fw1023_device_info(self):
        self.assertEqual('GW10K-ET', self.model_name)
        self.assertEqual('9010KETU000W0000', self.serial_number)
        self.assertEqual(10000, self.rated_power)
        self.assertEqual(2, self.modbus_version)
        self.assertEqual(254, self.ac_output_type)
        self.assertEqual(10, self.dsp1_version)
        self.assertEqual(10, self.dsp2_version)
        self.assertEqual(167, self.dsp_svn_version)
        self.assertEqual(23, self.arm_version)
        self.assertEqual(237, self.arm_svn_version)
        self.assertEqual('04029-10-S11', self.firmware)
        self.assertEqual('02041-23-S00', self.arm_firmware)

    def test_GW10K_ET_setting_fw1023(self):
        self.assertEqual(81, len(self.settings()))
        settings = {s.id_: s for s in self.settings()}
        self.assertEqual('PeakShavingMode', type(settings.get("peak_shaving_mode")).__name__)

    def test_GW10K_ET_runtime_data_fw1023(self):
        # Reset sensors
        self.loop.run_until_complete(self.read_device_info())
        self.sensor_map = {s.id_: s for s in self.sensors()}

        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(145, len(data))

        self.assertSensor('timestamp', datetime.strptime('2024-05-11 00:03:34', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 0.0, 'V', data)
        self.assertSensor('ipv1', 0.0, 'A', data)
        self.assertSensor('ppv1', 0, 'W', data)
        self.assertSensor('vpv2', 0.0, 'V', data)
        self.assertSensor('ipv2', 0.0, 'A', data)
        self.assertSensor('ppv2', 0, 'W', data)
        self.assertSensor('ppv', 0, 'W', data)
        self.assertSensor('pv2_mode', 0, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels not connected', '', data)
        self.assertSensor('pv1_mode', 0, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels not connected', '', data)
        self.assertSensor('vgrid', 244.4, 'V', data)
        self.assertSensor('igrid', 2.0, 'A', data)
        self.assertSensor('fgrid', 49.92, 'Hz', data)
        self.assertSensor('pgrid', 435, 'W', data)
        self.assertSensor('vgrid2', 244.3, 'V', data)
        self.assertSensor('igrid2', 1.2, 'A', data)
        self.assertSensor('fgrid2', 49.92, 'Hz', data)
        self.assertSensor('pgrid2', 241, 'W', data)
        self.assertSensor('vgrid3', 244.9, 'V', data)
        self.assertSensor('igrid3', 0.8, 'A', data)
        self.assertSensor('fgrid3', 49.92, 'Hz', data)
        self.assertSensor('pgrid3', 132, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 812, 'W', data)
        self.assertSensor('active_power', -11, 'W', data)
        self.assertSensor('grid_in_out', 0, '', data)
        self.assertSensor('grid_in_out_label', 'Idle', '', data)
        self.assertSensor('reactive_power', 0, 'var', data)
        self.assertSensor('apparent_power', 0, 'VA', data)
        self.assertSensor('backup_v1', 244.2, 'V', data)
        self.assertSensor('backup_i1', 1.3, 'A', data)
        self.assertSensor('backup_f1', 49.92, 'Hz', data)
        self.assertSensor('load_mode1', 1, '', data)
        self.assertSensor('backup_p1', 217, 'W', data)
        self.assertSensor('backup_v2', 243.8, 'V', data)
        self.assertSensor('backup_i2', 0.6, 'A', data)
        self.assertSensor('backup_f2', 49.92, 'Hz', data)
        self.assertSensor('load_mode2', 1, '', data)
        self.assertSensor('backup_p2', 114, 'W', data)
        self.assertSensor('backup_v3', 244.7, 'V', data)
        self.assertSensor('backup_i3', 0.3, 'A', data)
        self.assertSensor('backup_f3', 49.92, 'Hz', data)
        self.assertSensor('load_mode3', 1, '', data)
        self.assertSensor('backup_p3', 5, 'W', data)
        self.assertSensor('load_p1', 129, 'W', data)
        self.assertSensor('load_p2', 108, 'W', data)
        self.assertSensor('load_p3', 247, 'W', data)
        self.assertSensor('backup_ptotal', 321, 'W', data)
        self.assertSensor('load_ptotal', 502, 'W', data)
        self.assertSensor('ups_load', 5, '%', data)
        self.assertSensor('temperature_air', 36.4, 'C', data)
        self.assertSensor('temperature_module', 0.0, 'C', data)
        self.assertSensor('temperature', 32.5, 'C', data)
        self.assertSensor('function_bit', 16384, '', data)
        self.assertSensor('bus_voltage', 790.7, 'V', data)
        self.assertSensor('nbus_voltage', 395.1, 'V', data)
        self.assertSensor('vbattery1', 397.1, 'V', data)
        self.assertSensor('ibattery1', 2.0, 'A', data)
        self.assertSensor('pbattery1', 820, 'W', data)
        self.assertSensor('battery_mode', 2, '', data)
        self.assertSensor('battery_mode_label', 'Discharge', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor('e_total', 30630.9, 'kWh', data)
        self.assertSensor('e_day', 0, 'kWh', data)
        self.assertSensor('e_total_exp', 27208.5, 'kWh', data)
        self.assertSensor('h_total', 33055, 'h', data)
        self.assertSensor('e_day_exp', 0, 'kWh', data)
        self.assertSensor('e_total_imp', 70.3, 'kWh', data)
        self.assertSensor('e_day_imp', 0, 'kWh', data)
        self.assertSensor('e_load_total', 35366.4, 'kWh', data)
        self.assertSensor('e_load_day', 0, 'kWh', data)
        self.assertSensor('e_bat_charge_total', 9884.3, 'kWh', data)
        self.assertSensor('e_bat_charge_day', 0, 'kWh', data)
        self.assertSensor('e_bat_discharge_total', 8642.2, 'kWh', data)
        self.assertSensor('e_bat_discharge_day', 0.1, 'kWh', data)
        self.assertSensor('diagnose_result', 33554496, '', data)
        self.assertSensor('diagnose_result_label', 'Discharge Driver On, PF value set', '', data)
        self.assertSensor('house_consumption', 831, 'W', data)
        self.assertSensor('commode', 515, '', data)
        self.assertSensor('rssi', 1029, '', data)
        self.assertSensor('manufacture_code', 1543, '', data)


class GW6000_EH_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_RUNNING_DATA, 'GW6000_EH_running_data.hex')
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW6000_EH_device_info.hex')

    def test_GW6000_EH_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW6000-EH', self.model_name)
        self.assertEqual('00000EHU00000000', self.serial_number)
        self.assertEqual(6000, self.rated_power)
        self.assertEqual(0, self.modbus_version)
        self.assertEqual(254, self.ac_output_type)
        self.assertEqual(3, self.dsp1_version)
        self.assertEqual(3, self.dsp2_version)
        self.assertEqual(325, self.dsp_svn_version)
        self.assertEqual(16, self.arm_version)
        self.assertEqual(188, self.arm_svn_version)
        self.assertEqual('04034-03-S10', self.firmware)
        self.assertEqual('02041-16-S00', self.arm_firmware)

    def test_GW6000_EH_runtime_data(self):
        self.loop.run_until_complete(self.read_device_info())
        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(89, len(data))

        self.assertSensor('vpv1', 330.3, 'V', data)
        self.assertSensor('ipv1', 2.6, 'A', data)
        self.assertSensor('ppv1', 857, 'W', data)
        self.assertSensor('vpv2', 329.6, 'V', data)
        self.assertSensor('ipv2', 2.1, 'A', data)
        self.assertSensor('ppv2', 691, 'W', data)
        self.assertSensor('ppv', 1548, 'W', data)
        self.assertSensor('pv1_mode', 2, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv2_mode', 2, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('vgrid', 236.6, 'V', data)
        self.assertSensor('igrid', 6.6, 'A', data)
        self.assertSensor('fgrid', 49.97, 'Hz', data)
        self.assertSensor('pgrid', 1561, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 1561, 'W', data)
        self.assertSensor('active_power', -164, 'W', data)
        self.assertSensor('grid_in_out', 2, '', data)
        self.assertSensor('grid_in_out_label', 'Importing', '', data)
        self.assertSensor('reactive_power', -1, 'var', data)
        self.assertSensor('apparent_power', -1, 'VA', data)
        self.assertSensor('backup_v1', 0.0, 'V', data)
        self.assertSensor('backup_i1', 0.0, 'A', data)
        self.assertSensor('backup_f1', 0.0, 'Hz', data)
        self.assertSensor('load_mode1', 0, '', data)
        self.assertSensor('backup_p1', 0, 'W', data)
        self.assertSensor('load_p1', 1724, 'W', data)
        self.assertSensor('load_ptotal', 1725, 'W', data)
        self.assertSensor('backup_ptotal', 0, 'W', data)
        self.assertSensor('ups_load', 0, '%', data)
        self.assertSensor('temperature_air', 60.4, 'C', data)
        self.assertSensor('temperature_module', None, 'C', data)
        self.assertSensor('temperature', 38.6, 'C', data)
        self.assertSensor('function_bit', 256, '', data)
        self.assertSensor('bus_voltage', 380.6, 'V', data)
        self.assertSensor('nbus_voltage', 0, 'V', data)
        self.assertSensor('vbattery1', 0.0, 'V', data)
        self.assertSensor('ibattery1', 0.1, 'A', data)
        self.assertSensor('pbattery1', 0, 'W', data)
        self.assertSensor('battery_mode', 0, '', data)
        self.assertSensor('battery_mode_label', 'No battery', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 3, '', data)
        self.assertSensor('safety_country_label', 'ES-A', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor("e_total", 59.4, 'kWh', data)
        self.assertSensor("e_day", 22.0, 'kWh', data)
        self.assertSensor("e_total_exp", 58.6, 'kWh', data)
        self.assertSensor('h_total', 33, 'h', data)
        self.assertSensor("e_day_exp", 21.6, 'kWh', data)
        self.assertSensor("e_total_imp", 0, 'kWh', data)
        self.assertSensor("e_day_imp", 0, 'kWh', data)
        self.assertSensor("e_load_total", 70.1, 'kWh', data)
        self.assertSensor("e_load_day", 27.1, 'kWh', data)
        self.assertSensor("e_bat_charge_total", 0, 'kWh', data)
        self.assertSensor("e_bat_charge_day", 0, 'kWh', data)
        self.assertSensor("e_bat_discharge_total", 0, 'kWh', data)
        self.assertSensor("e_bat_discharge_day", 0, 'kWh', data)
        self.assertSensor('diagnose_result', 117983303, '', data)
        self.assertSensor('diagnose_result_label',
                          'Battery voltage low, Battery SOC low, Battery SOC in back, Discharge Driver On, Self-use load light, Battery Disconnected, Self-use off, Export power limit set, PF value set, Real power limit set',
                          '', data)
        self.assertSensor('house_consumption', 1712, 'W', data)


class GEH10_1U_10_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_RUNNING_DATA, 'GEH10-1U-10_running_data.hex')
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GEH10-1U-10_device_info.hex')

    def test_GEH10_1U_10_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('00000HSB00000000', self.serial_number)

    def test_GEH10_1U_10_runtime_data(self):
        # Reset sensors
        self.loop.run_until_complete(self.read_device_info())
        self.sensor_map = {s.id_: s for s in self.sensors()}

        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(125, len(data))

        self.assertSensor('timestamp', datetime.strptime('2023-01-26 11:34:07', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 242.3, 'V', data)
        self.assertSensor('ipv1', 11.5, 'A', data)
        self.assertSensor('ppv1', 2777, 'W', data)
        self.assertSensor('vpv2', 213.5, 'V', data)
        self.assertSensor('ipv2', 11.5, 'A', data)
        self.assertSensor('ppv2', 2455, 'W', data)
        self.assertSensor('vpv3', 333.3, 'V', data)
        self.assertSensor('ipv3', 11.0, 'A', data)
        self.assertSensor('ppv3', 3640, 'W', data)
        self.assertSensor('vpv4', 184.5, 'V', data)
        self.assertSensor('ipv4', 10.4, 'A', data)
        self.assertSensor('ppv4', 1915, 'W', data)
        self.assertSensor('ppv', 10787, 'W', data)
        self.assertSensor('pv4_mode', 2, '', data)
        self.assertSensor('pv4_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv3_mode', 2, '', data)
        self.assertSensor('pv3_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv2_mode', 2, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv1_mode', 2, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('vgrid', 242.9, 'V', data)
        self.assertSensor('igrid', 36.5, 'A', data)
        self.assertSensor('fgrid', 49.98, 'Hz', data)
        self.assertSensor('pgrid', 8710, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 8710, 'W', data)
        self.assertSensor('active_power', 4277, 'W', data)
        self.assertSensor('grid_in_out', 1, '', data)
        self.assertSensor('grid_in_out_label', 'Exporting', '', data)
        self.assertSensor('reactive_power', -1650, 'var', data)
        self.assertSensor('apparent_power', 8865, 'VA', data)
        self.assertSensor('backup_v1', 240.0, 'V', data)
        self.assertSensor('backup_i1', 0.7, 'A', data)
        self.assertSensor('backup_f1', 49.98, 'Hz', data)
        self.assertSensor('load_mode1', 1, '', data)
        self.assertSensor('backup_p1', 77, 'W', data)
        self.assertSensor('load_p1', 4356, 'W', data)
        self.assertSensor('backup_ptotal', 77, 'W', data)
        self.assertSensor('load_ptotal', 4356, 'W', data)
        self.assertSensor('ups_load', 1, '%', data)
        self.assertSensor('temperature_air', 0.0, 'C', data)
        self.assertSensor('temperature_module', -10.0, 'C', data)
        self.assertSensor('temperature', 67.0, 'C', data)
        self.assertSensor('function_bit', 257, '', data)
        self.assertSensor('bus_voltage', 458.4, 'V', data)
        self.assertSensor('nbus_voltage', 0, 'V', data)
        self.assertSensor('vbattery1', 406.1, 'V', data)
        self.assertSensor('ibattery1', -3.8, 'A', data)
        self.assertSensor('pbattery1', -1566, 'W', data)
        self.assertSensor('battery_mode', 3, '', data)
        self.assertSensor('battery_mode_label', 'Charge', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 9, '', data)
        self.assertSensor('safety_country_label', 'Australia A', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor('e_total', 10225.8, 'kWh', data)
        self.assertSensor('e_day', 23.1, 'kWh', data)
        self.assertSensor('e_total_exp', 10273.3, 'kWh', data)
        self.assertSensor('h_total', 3256, 'h', data)
        self.assertSensor('e_day_exp', 16.6, 'kWh', data)
        self.assertSensor('e_total_imp', 0, 'kWh', data)
        self.assertSensor('e_day_imp', 0, 'kWh', data)
        self.assertSensor('e_load_total', 4393.9, 'kWh', data)
        self.assertSensor('e_load_day', 10.7, 'kWh', data)
        self.assertSensor('e_bat_charge_total', 141.9, 'kWh', data)
        self.assertSensor('e_bat_charge_day', 9.6, 'kWh', data)
        self.assertSensor('e_bat_discharge_total', 117.5, 'kWh', data)
        self.assertSensor('e_bat_discharge_day', 2.6, 'kWh', data)
        self.assertSensor('diagnose_result', 33556864, '', data)
        self.assertSensor('diagnose_result_label',
                          'BMS: Discharge current low, APP: Discharge current too low, Self-use load light, PF value set',
                          '', data)
        self.assertSensor('house_consumption', 4944, 'W', data)
        self.assertSensor('battery_bms', 515, '', data)
        self.assertSensor('battery_index', 1029, '', data)
        self.assertSensor('battery_status', 1543, '', data)
        self.assertSensor('battery_temperature', 0.0, 'C', data)
        self.assertSensor('battery_charge_limit', 0, 'A', data)
        self.assertSensor('battery_discharge_limit', 0, 'A', data)
        self.assertSensor('battery_error_l', 0, '', data)
        self.assertSensor('battery_soc', 0, '%', data)
        self.assertSensor('battery_soh', 0, '%', data)
        self.assertSensor('battery_modules', 0, '', data)
        self.assertSensor('battery_warning_l', 0, '', data)
        self.assertSensor('battery_protocol', 0, '', data)
        self.assertSensor('battery_error_h', 0, '', data)
        self.assertSensor('battery_error', '', '', data)
        self.assertSensor('battery_warning_h', 0, '', data)
        self.assertSensor('battery_warning', '', '', data)
        self.assertSensor('battery_sw_version', 0, '', data)
        self.assertSensor('battery_hw_version', 0, '', data)
        self.assertSensor('battery_max_cell_temp_id', 0, '', data)
        self.assertSensor('battery_min_cell_temp_id', 0, '', data)
        self.assertSensor('battery_max_cell_voltage_id', 0, '', data)
        self.assertSensor('battery_min_cell_voltage_id', 0, '', data)
        self.assertSensor('battery_max_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery_min_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery_max_cell_voltage', 0.0, 'V', data)
        self.assertSensor('battery_min_cell_voltage', 0.0, 'V', data)
        self.assertSensor('commode', 515, '', data)
        self.assertSensor('rssi', 1029, '', data)
        self.assertSensor('manufacture_code', 1543, '', data)
        self.assertSensor('meter_test_status', 0, '', data)
        self.assertSensor('meter_comm_status', 0, '', data)
        self.assertSensor('active_power1', 0, 'W', data)
        self.assertSensor('active_power_total', 0, 'W', data)
        self.assertSensor('reactive_power_total', 0, 'var', data)
        self.assertSensor('meter_power_factor1', 0.0, '', data)
        self.assertSensor('meter_power_factor', 0.0, '', data)
        self.assertSensor('meter_freq', 0.0, 'Hz', data)
        self.assertSensor('meter_e_total_exp', 0.0, 'kWh', data)
        self.assertSensor('meter_e_total_imp', 0.0, 'kWh', data)
        self.assertSensor('meter_active_power1', 0, 'W', data)
        self.assertSensor('meter_active_power_total', 0, 'W', data)
        self.assertSensor('meter_reactive_power1', 0, 'var', data)
        self.assertSensor('meter_reactive_power_total', 0, 'var', data)
        self.assertSensor('meter_apparent_power1', 0, 'VA', data)
        self.assertSensor('meter_apparent_power_total', 0, 'VA', data)
        self.assertSensor('meter_type', 0, '', data)
        self.assertSensor('meter_sw_version', 0, '', data)

        self.assertFalse(self.sensor_map, f"Some sensors were not tested {self.sensor_map}")


class GW6000_ES_20_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW6000-ES-20_device_info.hex')

    def test_GW6000_ES_20_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW6000ES20', self.model_name)
        self.assertEqual('56000ESN00AW0000', self.serial_number)
        self.assertEqual(6050, self.rated_power)
        self.assertEqual(121, self.modbus_version)
        self.assertEqual(0, self.ac_output_type)
        self.assertEqual(2, self.dsp1_version)
        self.assertEqual(2, self.dsp2_version)
        self.assertEqual(3527, self.dsp_svn_version)
        self.assertEqual(5, self.arm_version)
        self.assertEqual(264, self.arm_svn_version)
        self.assertEqual('ffffffffffffffffffffffff', self.firmware)
        self.assertEqual('02020-05-S01', self.arm_firmware)


class GW25K_ET_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW25K-ET_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW25K-ET_running_data.hex')
        self.mock_response(self._READ_METER_DATA_EXTENDED2, ILLEGAL_DATA_ADDRESS)
        self.mock_response(self._READ_METER_DATA_EXTENDED, 'GW25K-ET_meter_data.hex')
        self.mock_response(self._READ_BATTERY_INFO, 'GW25K-ET_battery_info.hex')
        self.mock_response(self._READ_MPPT_DATA, 'GW25K-ET_mppt_data.hex')

    def test_GW25K_ET_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('', self.model_name)
        self.assertEqual('9025KETT00000000', self.serial_number)
        self.assertEqual(25000, self.rated_power)
        self.assertEqual(0, self.modbus_version)
        self.assertEqual(1, self.ac_output_type)
        self.assertEqual(6, self.dsp1_version)
        self.assertEqual(6, self.dsp2_version)
        self.assertEqual(6017, self.dsp_svn_version)
        self.assertEqual(8, self.arm_version)
        self.assertEqual(362, self.arm_svn_version)
        self.assertEqual('04062-', self.firmware)
        self.assertEqual('02020-08-S01', self.arm_firmware)

    def test_GW25K_ET_runtime_data(self):
        # Reset sensors
        self.loop.run_until_complete(self.read_device_info())

        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(237, len(data))

        self.sensor_map = {s.id_: s for s in self.sensors()}

        # self.assertEqual(36104, self.sensor_map.get("meter_e_total_exp").offset)

        self.assertSensor('timestamp', datetime.strptime('2023-12-03 14:07:07', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 737.9, 'V', data)
        self.assertSensor('ipv1', 1.4, 'A', data)
        self.assertSensor('ppv1', 1033, 'W', data)
        self.assertSensor('vpv2', 737.9, 'V', data)
        self.assertSensor('ipv2', 0.0, 'A', data)
        self.assertSensor('ppv2', 981, 'W', data)
        self.assertSensor('vpv3', 755.4, 'V', data)
        self.assertSensor('ipv3', 1.3, 'A', data)
        self.assertSensor('ppv3', 0, 'W', data)
        self.assertSensor('vpv4', 755.4, 'V', data)
        self.assertSensor('ipv4', 0.0, 'A', data)
        self.assertSensor('ppv4', 0, 'W', data)
        self.assertSensor('ppv', 2014, 'W', data)
        self.assertSensor('pv4_mode', 0, '', data)
        self.assertSensor('pv4_mode_label', 'PV panels not connected', '', data)
        self.assertSensor('pv3_mode', 0, '', data)
        self.assertSensor('pv3_mode_label', 'PV panels not connected', '', data)
        self.assertSensor('pv2_mode', 2, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv1_mode', 2, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('vgrid', 230.9, 'V', data)
        self.assertSensor('igrid', 2.9, 'A', data)
        self.assertSensor('fgrid', 49.95, 'Hz', data)
        self.assertSensor('pgrid', 589, 'W', data)
        self.assertSensor('vgrid2', 231.0, 'V', data)
        self.assertSensor('igrid2', 2.7, 'A', data)
        self.assertSensor('fgrid2', 49.97, 'Hz', data)
        self.assertSensor('pgrid2', 553, 'W', data)
        self.assertSensor('vgrid3', 230.4, 'V', data)
        self.assertSensor('igrid3', 4.2, 'A', data)
        self.assertSensor('fgrid3', 49.97, 'Hz', data)
        self.assertSensor('pgrid3', 803, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 1945, 'W', data)
        self.assertSensor('active_power', 1511, 'W', data)
        self.assertSensor('grid_in_out', 1, '', data)
        self.assertSensor('grid_in_out_label', 'Exporting', '', data)
        self.assertSensor('reactive_power', 1239, 'var', data)
        self.assertSensor('apparent_power', 2259, 'VA', data)
        self.assertSensor('backup_v1', 229.5, 'V', data)
        self.assertSensor('backup_i1', 3.0, 'A', data)
        self.assertSensor('backup_f1', 49.95, 'Hz', data)
        self.assertSensor('load_mode1', 0, '', data)
        self.assertSensor('backup_p1', 52, 'W', data)
        self.assertSensor('backup_v2', 230.0, 'V', data)
        self.assertSensor('backup_i2', 1.8, 'A', data)
        self.assertSensor('backup_f2', 49.97, 'Hz', data)
        self.assertSensor('load_mode2', 0, '', data)
        self.assertSensor('backup_p2', 15, 'W', data)
        self.assertSensor('backup_v3', 229.4, 'V', data)
        self.assertSensor('backup_i3', 3.2, 'A', data)
        self.assertSensor('backup_f3', 49.97, 'Hz', data)
        self.assertSensor('load_mode3', 0, '', data)
        self.assertSensor('backup_p3', 344, 'W', data)
        self.assertSensor('load_p1', 44, 'W', data)
        self.assertSensor('load_p2', 16, 'W', data)
        self.assertSensor('load_p3', 339, 'W', data)
        self.assertSensor('backup_ptotal', 408, 'W', data)
        self.assertSensor('load_ptotal', 26, 'W', data)
        self.assertSensor('ups_load', 7, '%', data)
        self.assertSensor('temperature_air', 46.2, 'C', data)
        self.assertSensor('temperature_module', 0.0, 'C', data)
        self.assertSensor('temperature', 43.0, 'C', data)
        self.assertSensor('function_bit', 0, '', data)
        self.assertSensor('bus_voltage', 773.3, 'V', data)
        self.assertSensor('nbus_voltage', 386.6, 'V', data)
        self.assertSensor('vbattery1', 215.2, 'V', data)
        self.assertSensor('ibattery1', 0.0, 'A', data)
        self.assertSensor('pbattery1', 0, 'W', data)
        self.assertSensor('battery_mode', 2, '', data)
        self.assertSensor('battery_mode_label', 'Discharge', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor('e_total', 160.3, 'kWh', data)
        self.assertSensor('e_day', 14.7, 'kWh', data)
        self.assertSensor('e_total_exp', 137.7, 'kWh', data)
        self.assertSensor('h_total', 388, 'h', data)
        self.assertSensor('e_day_exp', 2.9, 'kWh', data)
        self.assertSensor('e_total_imp', 14.8, 'kWh', data)
        self.assertSensor('e_day_imp', 1.0, 'kWh', data)
        self.assertSensor('e_load_total', 17.2, 'kWh', data)
        self.assertSensor('e_load_day', 0.2, 'kWh', data)
        self.assertSensor('e_bat_charge_total', 91.3, 'kWh', data)
        self.assertSensor('e_bat_charge_day', 11.0, 'kWh', data)
        self.assertSensor('e_bat_discharge_total', 69.6, 'kWh', data)
        self.assertSensor('e_bat_discharge_day', 0, 'kWh', data)
        self.assertSensor('diagnose_result', 33816960, '', data)
        self.assertSensor('diagnose_result_label',
                          'BMS: Discharge current low, APP: Discharge current too low, BMS: Charge disabled, PF value set',
                          '', data)
        self.assertSensor('house_consumption', 503, 'W', data)
        self.assertSensor('battery_bms', 255, '', data)
        self.assertSensor('battery_index', 311, '', data)
        self.assertSensor('battery_status', 1, '', data)
        self.assertSensor('battery_temperature', 23.0, 'C', data)
        self.assertSensor('battery_charge_limit', 0, 'A', data)
        self.assertSensor('battery_discharge_limit', 40, 'A', data)
        self.assertSensor('battery_error_l', 0, '', data)
        self.assertSensor('battery_soc', 100, '%', data)
        self.assertSensor('battery_soh', 100, '%', data)
        self.assertSensor('battery_modules', 4, '', data)
        self.assertSensor('battery_warning_l', 0, '', data)
        self.assertSensor('battery_protocol', 261, '', data)
        self.assertSensor('battery_error_h', 0, '', data)
        self.assertSensor('battery_error', '', '', data)
        self.assertSensor('battery_warning_h', 0, '', data)
        self.assertSensor('battery_warning', '', '', data)
        self.assertSensor('battery_sw_version', 790, '', data)
        self.assertSensor('battery_hw_version', 0, '', data)
        self.assertSensor('battery_max_cell_temp_id', 0, '', data)
        self.assertSensor('battery_min_cell_temp_id', 0, '', data)
        self.assertSensor('battery_max_cell_voltage_id', 0, '', data)
        self.assertSensor('battery_min_cell_voltage_id', 0, '', data)
        self.assertSensor('battery_max_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery_min_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery_max_cell_voltage', 0.0, 'V', data)
        self.assertSensor('battery_min_cell_voltage', 0.0, 'V', data)
        self.assertSensor('battery2_status', 515, '', data)
        self.assertSensor('battery2_temperature', 102.9, 'C', data)
        self.assertSensor('battery2_charge_limit', 1543, 'A', data)
        self.assertSensor('battery2_discharge_limit', 0, 'A', data)
        self.assertSensor('battery2_error_l', 0, '', data)
        self.assertSensor('battery2_soc', 0, '%', data)
        self.assertSensor('battery2_soh', 0, '%', data)
        self.assertSensor('battery2_modules', 0, '', data)
        self.assertSensor('battery2_warning_l', 0, '', data)
        self.assertSensor('battery2_protocol', 0, '', data)
        self.assertSensor('battery2_error_h', 0, '', data)
        self.assertSensor('battery2_error', '', '', data)
        self.assertSensor('battery2_warning_h', 0, '', data)
        self.assertSensor('battery2_warning', '', '', data)
        self.assertSensor('battery2_sw_version', 0, '', data)
        self.assertSensor('battery2_hw_version', 0, '', data)
        self.assertSensor('battery2_max_cell_temp_id', 0, '', data)
        self.assertSensor('battery2_min_cell_temp_id', 0, '', data)
        self.assertSensor('battery2_max_cell_voltage_id', 0, '', data)
        self.assertSensor('battery2_min_cell_voltage_id', 0, '', data)
        self.assertSensor('battery2_max_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery2_min_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery2_max_cell_voltage', 0.0, 'V', data)
        self.assertSensor('battery2_min_cell_voltage', 0.0, 'V', data)
        self.assertSensor('commode', 2, '', data)
        self.assertSensor('rssi', 100, '', data)
        self.assertSensor('manufacture_code', 10, '', data)
        self.assertSensor('meter_test_status', 273, '', data)
        self.assertSensor('meter_comm_status', 1, '', data)
        self.assertSensor('active_power1', -208, 'W', data)
        self.assertSensor('active_power2', -166, 'W', data)
        self.assertSensor('active_power3', -114, 'W', data)
        self.assertSensor('active_power_total', -489, 'W', data)
        self.assertSensor('reactive_power_total', 1197, 'var', data)
        self.assertSensor('meter_power_factor1', 24.236, '', data)
        self.assertSensor('meter_power_factor2', 32.336, '', data)
        self.assertSensor('meter_power_factor3', -28.2, '', data)
        self.assertSensor('meter_power_factor', 31.136, '', data)
        self.assertSensor('meter_freq', 49.99, 'Hz', data)
        self.assertSensor('meter_e_total_exp', 0.0, 'kWh', data)
        self.assertSensor('meter_e_total_imp', 0.0, 'kWh', data)
        self.assertSensor('meter_active_power1', -208, 'W', data)
        self.assertSensor('meter_active_power2', -166, 'W', data)
        self.assertSensor('meter_active_power3', -114, 'W', data)
        self.assertSensor('meter_active_power_total', -489, 'W', data)
        self.assertSensor('meter_reactive_power1', 421, 'var', data)
        self.assertSensor('meter_reactive_power2', 440, 'var', data)
        self.assertSensor('meter_reactive_power3', 335, 'var', data)
        self.assertSensor('meter_reactive_power_total', 1197, 'var', data)
        self.assertSensor('meter_apparent_power1', -511, 'VA', data)
        self.assertSensor('meter_apparent_power2', -509, 'VA', data)
        self.assertSensor('meter_apparent_power3', -448, 'VA', data)
        self.assertSensor('meter_apparent_power_total', -1470, 'VA', data)
        self.assertSensor('meter_type', 2, '', data)
        self.assertSensor('meter_sw_version', 5, '', data)
        self.assertSensor('meter2_active_power', 0, 'W', data)
        self.assertSensor('meter2_e_total_exp', 0.0, 'kWh', data)
        self.assertSensor('meter2_e_total_imp', 0.0, 'kWh', data)
        self.assertSensor('meter2_comm_status', 0, '', data)
        self.assertSensor('meter_voltage1', 229.0, 'V', data)
        self.assertSensor('meter_voltage2', 229.6, 'V', data)
        self.assertSensor('meter_voltage3', 228.8, 'V', data)
        self.assertSensor('meter_current1', 2.2, 'A', data)
        self.assertSensor('meter_current2', 2.2, 'A', data)
        self.assertSensor('meter_current3', 1.9, 'A', data)
        self.assertSensor('ppv_total', 529, 'W', data)
        self.assertSensor('pv_channel', 2, '', data)
        self.assertSensor('vpv5', 0.0, 'V', data)
        self.assertSensor('ipv5', 0.0, 'A', data)
        self.assertSensor('vpv6', 0.0, 'V', data)
        self.assertSensor('ipv6', 0.0, 'A', data)
        self.assertSensor('vpv7', 0.0, 'V', data)
        self.assertSensor('ipv7', 0.0, 'A', data)
        self.assertSensor('vpv8', 0.0, 'V', data)
        self.assertSensor('ipv8', 0.0, 'A', data)
        self.assertSensor('vpv9', 0.0, 'V', data)
        self.assertSensor('ipv9', 0.0, 'A', data)
        self.assertSensor('vpv10', 0.0, 'V', data)
        self.assertSensor('ipv10', 0.0, 'A', data)
        self.assertSensor('vpv11', 0.0, 'V', data)
        self.assertSensor('ipv11', 0.0, 'A', data)
        self.assertSensor('vpv12', 0.0, 'V', data)
        self.assertSensor('ipv12', 0.0, 'A', data)
        self.assertSensor('vpv13', 0.0, 'V', data)
        self.assertSensor('ipv13', 0.0, 'A', data)
        self.assertSensor('vpv14', 0.0, 'V', data)
        self.assertSensor('ipv14', 0.0, 'A', data)
        self.assertSensor('vpv15', 0.0, 'V', data)
        self.assertSensor('ipv15', 0.0, 'A', data)
        self.assertSensor('vpv16', 0.0, 'V', data)
        self.assertSensor('ipv16', 0.0, 'A', data)
        self.assertSensor('pmppt1', 232, 'W', data)
        self.assertSensor('pmppt2', 299, 'W', data)
        self.assertSensor('pmppt3', 0, 'W', data)
        self.assertSensor('pmppt4', 0, 'W', data)
        self.assertSensor('pmppt5', 0, 'W', data)
        self.assertSensor('pmppt6', 0, 'W', data)
        self.assertSensor('pmppt7', 0, 'W', data)
        self.assertSensor('pmppt8', 0, 'W', data)
        self.assertSensor('imppt1', 0.3, 'A', data)
        self.assertSensor('imppt2', 0.4, 'A', data)
        self.assertSensor('imppt3', 0, 'A', data)
        self.assertSensor('imppt4', 0, 'A', data)
        self.assertSensor('imppt5', 0, 'A', data)
        self.assertSensor('imppt6', 0, 'A', data)
        self.assertSensor('imppt7', 0, 'A', data)
        self.assertSensor('imppt8', 0, 'A', data)
        self.assertSensor('reactive_power1', 0, 'var', data)
        self.assertSensor('reactive_power2', 0, 'var', data)
        self.assertSensor('reactive_power3', 0, 'var', data)
        self.assertSensor('apparent_power1', 0, 'VA', data)
        self.assertSensor('apparent_power2', 0, 'VA', data)
        self.assertSensor('apparent_power3', 0, 'VA', data)

        self.assertFalse(self.sensor_map, f"Some sensors were not tested {self.sensor_map}")


class GW29K9_ET_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW29K9-ET_device_info.hex')
        self.mock_response(self._READ_RUNNING_DATA, 'GW29K9-ET_running_data.hex')
        self.mock_response(self._READ_METER_DATA_EXTENDED2, ILLEGAL_DATA_ADDRESS)
        self.mock_response(self._READ_METER_DATA_EXTENDED, 'GW29K9-ET_meter_data.hex')
        self.mock_response(self._READ_BATTERY_INFO, 'GW29K9-ET_battery_info.hex')
        self.mock_response(self._READ_BATTERY2_INFO, 'GW29K9-ET_battery2_info.hex')
        self.mock_response(self._READ_MPPT_DATA, 'GW29K9-ET_mppt_data.hex')

    def test_GW29K9_ET_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('', self.model_name)
        self.assertEqual('929K9ETT00CW0000', self.serial_number)
        self.assertEqual(29900, self.rated_power)
        self.assertEqual(0, self.modbus_version)
        self.assertEqual(1, self.ac_output_type)
        self.assertEqual(2, self.dsp1_version)
        self.assertEqual(2, self.dsp2_version)
        self.assertEqual(79, self.dsp_svn_version)
        self.assertEqual(3, self.arm_version)
        self.assertEqual(67, self.arm_svn_version)
        self.assertEqual('04062-', self.firmware)
        self.assertEqual('02020-03-S01', self.arm_firmware)

    def test_GW29K9_ET_runtime_data(self):
        # Reset sensors
        self.loop.run_until_complete(self.read_device_info())

        data = self.loop.run_until_complete(self.read_runtime_data())
        self.assertEqual(211, len(data))

        self.sensor_map = {s.id_: s for s in self.sensors()}

        self.assertSensor('timestamp', datetime.strptime('2024-01-17 14:49:14', '%Y-%m-%d %H:%M:%S'), '', data)
        self.assertSensor('vpv1', 682.9, 'V', data)
        self.assertSensor('ipv1', 1.5, 'A', data)
        self.assertSensor('ppv1', 478, 'W', data)
        self.assertSensor('vpv2', 682.9, 'V', data)
        self.assertSensor('ipv2', 0.0, 'A', data)
        self.assertSensor('ppv2', 679, 'W', data)
        self.assertSensor('vpv3', 577.3, 'V', data)
        self.assertSensor('ipv3', 1.8, 'A', data)
        self.assertSensor('ppv3', 390, 'W', data)
        self.assertSensor('vpv4', 577.3, 'V', data)
        self.assertSensor('ipv4', 0.0, 'A', data)
        self.assertSensor('ppv4', 0, 'W', data)
        self.assertSensor('ppv', 1547, 'W', data)
        self.assertSensor('pv4_mode', 0, '', data)
        self.assertSensor('pv4_mode_label', 'PV panels not connected', '', data)
        self.assertSensor('pv3_mode', 2, '', data)
        self.assertSensor('pv3_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv2_mode', 2, '', data)
        self.assertSensor('pv2_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('pv1_mode', 2, '', data)
        self.assertSensor('pv1_mode_label', 'PV panels connected, producing power', '', data)
        self.assertSensor('vgrid', 231.3, 'V', data)
        self.assertSensor('igrid', 2.9, 'A', data)
        self.assertSensor('fgrid', 49.99, 'Hz', data)
        self.assertSensor('pgrid', 583, 'W', data)
        self.assertSensor('vgrid2', 232.9, 'V', data)
        self.assertSensor('igrid2', 2.7, 'A', data)
        self.assertSensor('fgrid2', 49.99, 'Hz', data)
        self.assertSensor('pgrid2', 565, 'W', data)
        self.assertSensor('vgrid3', 233.6, 'V', data)
        self.assertSensor('igrid3', 2.9, 'A', data)
        self.assertSensor('fgrid3', 49.97, 'Hz', data)
        self.assertSensor('pgrid3', 587, 'W', data)
        self.assertSensor('grid_mode', 1, '', data)
        self.assertSensor('grid_mode_label', 'Connected to grid', '', data)
        self.assertSensor('total_inverter_power', 1735, 'W', data)
        self.assertSensor('active_power', -5403, 'W', data)
        self.assertSensor('grid_in_out', 2, '', data)
        self.assertSensor('grid_in_out_label', 'Importing', '', data)
        self.assertSensor('reactive_power', 307, 'var', data)
        self.assertSensor('apparent_power', 1975, 'VA', data)
        self.assertSensor('backup_v1', 229.9, 'V', data)
        self.assertSensor('backup_i1', 0.7, 'A', data)
        self.assertSensor('backup_f1', 49.98, 'Hz', data)
        self.assertSensor('load_mode1', 0, '', data)
        self.assertSensor('backup_p1', 21, 'W', data)
        self.assertSensor('backup_v2', 231.5, 'V', data)
        self.assertSensor('backup_i2', 0.7, 'A', data)
        self.assertSensor('backup_f2', 50.0, 'Hz', data)
        self.assertSensor('load_mode2', 0, '', data)
        self.assertSensor('backup_p2', 5, 'W', data)
        self.assertSensor('backup_v3', 232.9, 'V', data)
        self.assertSensor('backup_i3', 0.6, 'A', data)
        self.assertSensor('backup_f3', 49.97, 'Hz', data)
        self.assertSensor('load_mode3', 0, '', data)
        self.assertSensor('backup_p3', 37, 'W', data)
        self.assertSensor('load_p1', 647, 'W', data)
        self.assertSensor('load_p2', 640, 'W', data)
        self.assertSensor('load_p3', 651, 'W', data)
        self.assertSensor('backup_ptotal', 66, 'W', data)
        self.assertSensor('load_ptotal', 7072, 'W', data)
        self.assertSensor('ups_load', 1, '%', data)
        self.assertSensor('temperature_air', 24.1, 'C', data)
        self.assertSensor('temperature_module', 0.0, 'C', data)
        self.assertSensor('temperature', 20.5, 'C', data)
        self.assertSensor('function_bit', 0, '', data)
        self.assertSensor('bus_voltage', 760.4, 'V', data)
        self.assertSensor('nbus_voltage', 380.2, 'V', data)
        self.assertSensor('vbattery1', 0.0, 'V', data)
        self.assertSensor('ibattery1', -0.1, 'A', data)
        self.assertSensor('pbattery1', 0, 'W', data)
        self.assertSensor('battery_mode', 0, '', data)
        self.assertSensor('battery_mode_label', 'No battery', '', data)
        self.assertSensor('warning_code', 0, '', data)
        self.assertSensor('safety_country', 32, '', data)
        self.assertSensor('safety_country_label', '50Hz 230Vac Default', '', data)
        self.assertSensor('work_mode', 1, '', data)
        self.assertSensor('work_mode_label', 'Normal (On-Grid)', '', data)
        self.assertSensor('operation_mode', 0, '', data)
        self.assertSensor('error_codes', 0, '', data)
        self.assertSensor('errors', '', '', data)
        self.assertSensor('e_total', 4562.3, 'kWh', data)
        self.assertSensor('e_day', 0.9, 'kWh', data)
        self.assertSensor('e_total_exp', 4489.7, 'kWh', data)
        self.assertSensor('h_total', 1175, 'h', data)
        self.assertSensor('e_day_exp', 1.2, 'kWh', data)
        self.assertSensor('e_total_imp', 8.7, 'kWh', data)
        self.assertSensor('e_day_imp', 0, 'kWh', data)
        self.assertSensor('e_load_total', 10742.2, 'kWh', data)
        self.assertSensor('e_load_day', 43.8, 'kWh', data)
        self.assertSensor('e_bat_charge_total', 0, 'kWh', data)
        self.assertSensor('e_bat_charge_day', 0, 'kWh', data)
        self.assertSensor('e_bat_discharge_total', 0, 'kWh', data)
        self.assertSensor('e_bat_discharge_day', 0, 'kWh', data)
        self.assertSensor('diagnose_result', 33816782, '', data)
        self.assertSensor('diagnose_result_label',
                          'Battery SOC low, Battery SOC in back, BMS: Discharge disabled, '
                          'Discharge Driver On, BMS: Discharge current low, BMS: Charge disabled, PF value set',
                          '', data)
        self.assertSensor('house_consumption', 6950, 'W', data)
        self.assertSensor('commode', 0, '', data)
        self.assertSensor('rssi', 100, '', data)
        self.assertSensor('manufacture_code', 10, '', data)
        self.assertSensor('meter_test_status', 0, '', data)
        self.assertSensor('meter_comm_status', 1, '', data)
        self.assertSensor('active_power1', -599, 'W', data)
        self.assertSensor('active_power2', -567, 'W', data)
        self.assertSensor('active_power3', -2911, 'W', data)
        self.assertSensor('active_power_total', -4077, 'W', data)
        self.assertSensor('reactive_power_total', 2831, 'var', data)
        self.assertSensor('meter_power_factor1', -6.38, '', data)
        self.assertSensor('meter_power_factor2', -4.95, '', data)
        self.assertSensor('meter_power_factor3', -9.17, '', data)
        self.assertSensor('meter_power_factor', -7.46, '', data)
        self.assertSensor('meter_freq', 49.99, 'Hz', data)
        self.assertSensor('meter_e_total_exp', 0.0, 'kWh', data)
        self.assertSensor('meter_e_total_imp', 0.0, 'kWh', data)
        self.assertSensor('meter_active_power1', -599, 'W', data)
        self.assertSensor('meter_active_power2', -567, 'W', data)
        self.assertSensor('meter_active_power3', -2911, 'W', data)
        self.assertSensor('meter_active_power_total', -4077, 'W', data)
        self.assertSensor('meter_reactive_power1', 647, 'var', data)
        self.assertSensor('meter_reactive_power2', 1105, 'var', data)
        self.assertSensor('meter_reactive_power3', 1077, 'var', data)
        self.assertSensor('meter_reactive_power_total', 2831, 'var', data)
        self.assertSensor('meter_apparent_power1', -1076, 'VA', data)
        self.assertSensor('meter_apparent_power2', -1414, 'VA', data)
        self.assertSensor('meter_apparent_power3', -3175, 'VA', data)
        self.assertSensor('meter_apparent_power_total', -5667, 'VA', data)
        self.assertSensor('meter_type', 2, '', data)
        self.assertSensor('meter_sw_version', 0, '', data)
        self.assertSensor('meter2_active_power', 0, 'W', data)
        self.assertSensor('meter2_e_total_exp', 0.0, 'kWh', data)
        self.assertSensor('meter2_e_total_imp', 0.0, 'kWh', data)
        self.assertSensor('meter2_comm_status', 0, '', data)
        self.assertSensor('meter_voltage1', 231.1, 'V', data)
        self.assertSensor('meter_voltage2', 233.2, 'V', data)
        self.assertSensor('meter_voltage3', 232.6, 'V', data)
        self.assertSensor('meter_current1', 4.6, 'A', data)
        self.assertSensor('meter_current2', 6.0, 'A', data)
        self.assertSensor('meter_current3', 13.6, 'A', data)
        self.assertSensor('battery2_status', 0, '', data)
        self.assertSensor('battery2_temperature', 0.0, 'C', data)
        self.assertSensor('battery2_charge_limit', 0, 'A', data)
        self.assertSensor('battery2_discharge_limit', 0, 'A', data)
        self.assertSensor('battery2_error_l', 4096, '', data)
        self.assertSensor('battery2_soc', 0, '%', data)
        self.assertSensor('battery2_soh', 0, '%', data)
        self.assertSensor('battery2_modules', 6, '', data)
        self.assertSensor('battery2_warning_l', 0, '', data)
        self.assertSensor('battery2_protocol', 288, '', data)
        self.assertSensor('battery2_error_h', 0, '', data)
        self.assertSensor('battery2_error', '', '', data)
        self.assertSensor('battery2_warning_h', 0, '', data)
        self.assertSensor('battery2_warning', '', '', data)
        self.assertSensor('battery2_sw_version', 0, '', data)
        self.assertSensor('battery2_hw_version', 0, '', data)
        self.assertSensor('battery2_max_cell_temp_id', 0, '', data)
        self.assertSensor('battery2_min_cell_temp_id', 0, '', data)
        self.assertSensor('battery2_max_cell_voltage_id', 0, '', data)
        self.assertSensor('battery2_min_cell_voltage_id', 0, '', data)
        self.assertSensor('battery2_max_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery2_min_cell_temp', 0.0, 'C', data)
        self.assertSensor('battery2_max_cell_voltage', 0.0, 'V', data)
        self.assertSensor('battery2_min_cell_voltage', 0.0, 'V', data)
        self.assertSensor('ppv_total', 1565, 'W', data)
        self.assertSensor('pv_channel', 3, '', data)
        self.assertSensor('vpv5', 301.5, 'V', data)
        self.assertSensor('ipv5', 2.0, 'A', data)
        self.assertSensor('vpv6', 301.5, 'V', data)
        self.assertSensor('ipv6', 0.0, 'A', data)
        self.assertSensor('vpv7', 0.0, 'V', data)
        self.assertSensor('ipv7', 0.0, 'A', data)
        self.assertSensor('vpv8', 0.0, 'V', data)
        self.assertSensor('ipv8', 0.0, 'A', data)
        self.assertSensor('vpv9', 0.0, 'V', data)
        self.assertSensor('ipv9', 0.0, 'A', data)
        self.assertSensor('vpv10', 0.0, 'V', data)
        self.assertSensor('ipv10', 0.0, 'A', data)
        self.assertSensor('vpv11', 0.0, 'V', data)
        self.assertSensor('ipv11', 0.0, 'A', data)
        self.assertSensor('vpv12', 0.0, 'V', data)
        self.assertSensor('ipv12', 0.0, 'A', data)
        self.assertSensor('vpv13', 0.0, 'V', data)
        self.assertSensor('ipv13', 0.0, 'A', data)
        self.assertSensor('vpv14', 0.0, 'V', data)
        self.assertSensor('ipv14', 0.0, 'A', data)
        self.assertSensor('vpv15', 0.0, 'V', data)
        self.assertSensor('ipv15', 0.0, 'A', data)
        self.assertSensor('vpv16', 0.0, 'V', data)
        self.assertSensor('ipv16', 0.0, 'A', data)
        self.assertSensor('pmppt1', 471, 'W', data)
        self.assertSensor('pmppt2', 729, 'W', data)
        self.assertSensor('pmppt3', 365, 'W', data)
        self.assertSensor('pmppt4', 0, 'W', data)
        self.assertSensor('pmppt5', 0, 'W', data)
        self.assertSensor('pmppt6', 0, 'W', data)
        self.assertSensor('pmppt7', 0, 'W', data)
        self.assertSensor('pmppt8', 0, 'W', data)
        self.assertSensor('imppt1', 0.7, 'A', data)
        self.assertSensor('imppt2', 1.3, 'A', data)
        self.assertSensor('imppt3', 1.3, 'A', data)
        self.assertSensor('imppt4', 0.0, 'A', data)
        self.assertSensor('imppt5', 0.0, 'A', data)
        self.assertSensor('imppt6', 0.0, 'A', data)
        self.assertSensor('imppt7', 0.0, 'A', data)
        self.assertSensor('imppt8', 0.0, 'A', data)
        self.assertSensor('reactive_power1', 0, 'var', data)
        self.assertSensor('reactive_power2', 0, 'var', data)
        self.assertSensor('reactive_power3', 0, 'var', data)
        self.assertSensor('apparent_power1', 0, 'VA', data)
        self.assertSensor('apparent_power2', 0, 'VA', data)
        self.assertSensor('apparent_power3', 0, 'VA', data)

        self.assertFalse(self.sensor_map, f"Some sensors were not tested {self.sensor_map}")


class GW5K_BT_Test(EtMock):

    def __init__(self, methodName='runTest'):
        EtMock.__init__(self, methodName)
        self.mock_response(self._READ_DEVICE_VERSION_INFO, 'GW5K-BT_device_info.hex')
        self.mock_response(ModbusRtuReadCommand(0xf7, 47547, 6), 'NO RESPONSE')

    def test_GW5K_BT_device_info(self):
        self.loop.run_until_complete(self.read_device_info())
        self.assertEqual('GW5K-BT', self.model_name)
        self.assertEqual('95000BTU203W0000', self.serial_number)
        self.assertEqual(5000, self.rated_power)
        self.assertEqual(0, self.modbus_version)
        self.assertEqual(254, self.ac_output_type)
        self.assertEqual(3, self.dsp1_version)
        self.assertEqual(3, self.dsp2_version)
        self.assertEqual(124, self.dsp_svn_version)
        self.assertEqual(11, self.arm_version)
        self.assertEqual(147, self.arm_svn_version)
        self.assertEqual('04029-03-S10', self.firmware)
        self.assertEqual('02041-11-S00', self.arm_firmware)
